#!/usr/bin/env python

# This file is part of Androguard.
#
# Copyright (C) 2012, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

import sys, os

from optparse import OptionParser

from androguard.core.androgen import Androguard
from androguard.core import androconf
from androguard.core.analysis import analysis
from androguard.core.bytecode import method2dot, method2format

option_0 = { 'name' : ('-i', '--input'), 'help' : 'file : use this filename', 'nargs' : 1 }
option_1 = { 'name' : ('-o', '--output'), 'help' : 'base directory to output all files', 'nargs' : 1 }

option_2 = { 'name' : ('-d', '--dot'), 'help' : 'write the method in dot format', 'action' : 'count' }
option_3 = { 'name' : ('-f', '--format'), 'help' : 'write the method in specific format (png, ...)', 'nargs' : 1 }

option_4 = { 'name' : ('-v', '--version'), 'help' : 'version of the API', 'action' : 'count' }

options = [option_0, option_1, option_2, option_3, option_4]

def valid_class_name( class_name ):
    if class_name[-1] == ";" :
        return class_name[1:-1]
    return class_name

def create_directory( class_name, output ) :
    output_name = output
    if output_name[-1] != "/" :
        output_name = output_name + "/"

    try :
        os.makedirs( output_name + class_name )
    except OSError :
        pass

def create_directories( a, output ) :
    for vm in a.get_vms() :
        for class_name in vm.get_classes_names() :
            create_directory( valid_class_name( class_name ), output )

def export_apps_to_format( a, output, dot=None, _format=None ) :
    output_name = output
    if output_name[-1] != "/" :
        output_name = output_name + "/"

    for vm in a.get_vms() :
        x = analysis.VMAnalysis( vm )
        for method in vm.get_methods() :
            filename = output_name + valid_class_name( method.get_class_name() )
            if filename[-1] != "/" :
                filename = filename + "/"

            descriptor = method.get_descriptor()
            descriptor = descriptor.replace(";", "")
            descriptor = descriptor.replace(" ", "")
            descriptor = descriptor.replace("(", "-")
            descriptor = descriptor.replace(")", "-")
            descriptor = descriptor.replace("/", "_")

            filename = filename + method.get_name() + descriptor


            buff = method2dot( x.get_method( method ) )

            if dot :
                fd = open( filename + ".dot", "w")
                fd.write( buff )
                fd.close()

            if _format :
                method2format( filename + "." + _format, _format, raw = buff )

def main(options, arguments) :
    if options.input != None and options.output != None :
        a = Androguard( [ options.input ] )

        if options.dot != None or options.format != None :
            create_directories( a, options.output )
            export_apps_to_format( a, options.output, options.dot, options.format )
        else :
          print "Please, specify a format or dot option"

    elif options.version != None :
        print "Androdd version %s" % androconf.ANDROGUARD_VERSION
    
    else :
      print "Please, specify an input file and an output directory"

if __name__ == "__main__" :
    parser = OptionParser()
    for option in options :
        param = option['name']
        del option['name']
        parser.add_option(*param, **option)

    options, arguments = parser.parse_args()
    sys.argv[:] = arguments
    main(options, arguments)



# This file is part of Androguard.
#
# Copyright (c) 2012 Geoffroy Gueguen <geoffroy.gueguen@gmail.com>
# All Rights Reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

from dad.util import get_type, log, ACCESS_FLAGS_METHODS
from dad.opcode_ins import Op
from dad.instruction import Constant, ThisParam, BinaryCompExpression


class Writer(object):
    def __init__(self, graph, method):
        self.graph = graph
        self.method = method
        self.visited_nodes = set()
        self.ind = 4
        self.buffer = []
        self.loop_follow = [None]
        self.latch_node = [None]
        self.if_follow = [None]
        self.switch_follow = [None]
        self.next_case = None
        self.skip = False
        self.need_break = True

    def __str__(self):
        return ''.join(self.buffer)

    def inc_ind(self, i=1):
        self.ind += (4 * i)

    def dec_ind(self, i=1):
        self.ind -= (4 * i)

    def space(self):
        if self.skip:
            self.skip = False
            return ''
        return ' ' * self.ind

    def write_ind(self):
        if self.skip:
            self.skip = False
        else:
            self.write(self.space())

    def write(self, s):
        self.buffer.append(s)

    def end_ins(self):
        self.write(';\n')

    def visit_ins(self, ins):
        ins.visit(self)

    def write_method(self):
        acc = []
        access = self.method.access
        self.constructor = 0x10000 in access
        for i in self.method.access:
            if i == 0x10000:
                continue
            acc.append(ACCESS_FLAGS_METHODS.get(i))
        if self.constructor:
            name = get_type(self.method.method.class_name).split('.')[-1]
            proto = '%s %s(' % (' '.join(acc), name)
        else:
            name = self.method.name
            proto = '%s %s %s(' % (' '.join(acc), self.method.type, name)
        self.write('%s%s' % (self.space(), proto))
        if 0x8 in self.method.access:
            params = self.method.lparams
        else:
            params = self.method.lparams[1:]
        proto = ''
        if self.method.params_type:
            proto = ', '.join(['%s p%s' % (get_type(p_type), param) for
                        p_type, param in zip(self.method.params_type, params)])
        self.write('%s)' % proto)
        if self.graph is None:
            return self.write(';')
        self.write('\n%s{\n' % self.space())
        self.inc_ind()
#        for v, var in self.method.var_to_name.iteritems():
#            var.visit_decl(self)
        self.visit_node(self.graph.get_entry())
        self.dec_ind()
        self.write('%s}\n' % self.space())

    def visit_node(self, node):
        if node in (self.if_follow[-1], self.switch_follow[-1],
                    self.loop_follow[-1], self.latch_node[-1]):
            return
        if node in self.visited_nodes:
            return
        self.visited_nodes.add(node)
        node.visit(self)

    def visit_loop_node(self, loop):
        follow = loop.get_loop_follow()
        if follow is None and not loop.looptype.endless():
            log('Loop has no follow !', 'error')
        if loop.looptype.pretest():
            if loop.true is follow:
                loop.neg()
                loop.true, loop.false = loop.false, loop.true
            self.write('%swhile (' % self.space())
            loop.visit_cond(self)
            self.write(') {\n')
        elif loop.looptype.posttest():
            self.write('%sdo {\n' % self.space())
            self.latch_node.append(loop.latch)
        elif loop.looptype.endless():
            self.write('%swhile(true) {\n' % self.space())
        self.inc_ind()
        self.loop_follow.append(follow)
        if loop.looptype.pretest():
            self.visit_node(loop.true)
        else:
            self.visit_node(loop.cond)
        self.loop_follow.pop()
        self.dec_ind()
        if loop.looptype.pretest():
            self.write('%s}\n' % self.space())
        elif loop.looptype.posttest():
            self.latch_node.pop()
            self.write('%s} while(' % self.space())
            loop.latch.visit_cond(self)
            self.write(');\n')
        else:
            self.inc_ind()
            self.visit_node(loop.latch)
            self.dec_ind()
            self.write('%s}\n' % self.space())
        if follow is not None:
            self.visit_node(follow)

    def visit_cond_node(self, cond):
        follow = cond.get_if_follow()
        if cond.false is self.loop_follow[-1]:
            cond.neg()
            cond.true, cond.false = cond.false, cond.true
            self.write('%sif(' % self.space())
            cond.visit_cond(self)
            self.write(') {\n')
            self.inc_ind()
            self.write('%sbreak;\n' % self.space())
            self.dec_ind()
            self.write('%s}\n' % self.space())
            self.visit_node(cond.false)
        elif follow is not None:
            is_else = not (follow in (cond.true, cond.false))
            if (cond.true in (follow, self.next_case)
                                                or cond.num > cond.true.num):
                cond.neg()
                cond.true, cond.false = cond.false, cond.true
            self.if_follow.append(follow)
            if not cond.true in self.visited_nodes:
                self.write('%sif(' % self.space())
                cond.visit_cond(self)
                self.write(') {\n')
                self.inc_ind()
                self.visit_node(cond.true)
                self.dec_ind()
            if is_else and not cond.false in self.visited_nodes:
                self.write('%s} else {\n' % self.space())
                self.inc_ind()
                self.visit_node(cond.false)
                self.dec_ind()
            self.if_follow.pop()
            self.write('%s}\n' % self.space())
            self.visit_node(follow)
        else:
            self.write('%sif (' % self.space())
            cond.visit_cond(self)
            self.write(') {\n')
            self.inc_ind()
            self.visit_node(cond.true)
            self.dec_ind()
            self.write('%s} else {\n' % self.space())
            self.inc_ind()
            self.visit_node(cond.false)
            self.dec_ind()
            self.write('%s}\n' % self.space())

    def visit_short_circuit_condition(self, nnot, aand, cond1, cond2):
        if nnot:
            cond1.neg()
        self.write('(')
        cond1.visit_cond(self)
        self.write(') %s (' % ['||', '&&'][aand])
        cond2.visit_cond(self)
        self.write(')')

    def visit_switch_node(self, switch):
        lins = switch.get_ins()
        for ins in lins[:-1]:
            self.visit_ins(ins)
        switch_ins = switch.get_ins()[-1]
        self.write('%sswitch(' % self.space())
        self.visit_ins(switch_ins)
        self.write(') {\n')
        follow = switch.switch_follow
        cases = switch.cases
        self.switch_follow.append(follow)
        default = switch.default
        for i, node in enumerate(cases):
            if node in self.visited_nodes:
                continue
            self.inc_ind()
            for case in switch.node_to_case[node]:
                self.write('%scase %d:\n' % (self.space(), case))
            if i + 1 < len(cases):
                self.next_case = cases[i + 1]
            else:
                self.next_case = None
            if node is default:
                self.write('%sdefault:\n' % self.space())
                default = None
            self.inc_ind()
            self.visit_node(node)
            if self.need_break:
                self.write('%sbreak;\n' % self.space())
            else:
                self.need_break = True
            self.dec_ind(2)
        if default not in (None, follow):
            self.inc_ind()
            self.write('%sdefault:\n' % self.space())
            self.inc_ind()
            self.visit_node(default)
            self.dec_ind(2)
        self.write('%s}\n' % self.space())
        self.switch_follow.pop()
        self.visit_node(follow)

    def visit_statement_node(self, stmt):
        sucs = self.graph.sucs(stmt)
        for ins in stmt.get_ins():
            self.visit_ins(ins)
        if len(sucs) == 0:
            return
        follow = sucs[0]
        self.visit_node(follow)

    def visit_return_node(self, ret):
        self.need_break = False
        for ins in ret.get_ins():
            self.visit_ins(ins)

    def visit_throw_node(self, throw):
        for ins in throw.get_ins():
            self.visit_ins(ins)

#    def visit_decl(self, var):
#        self.write('%sdecl v%s' % (SPACE * self.ind, var))
#        self.end_ins()

    def visit_constant(self, cst):
        if isinstance(cst, str):
            return self.write(string('%s' % cst))
        self.write('%s' % cst)

    def visit_base_class(self, cls):
        self.write(cls)

    def visit_variable(self, var):
        if isinstance(var, str):
            return self.write(var)
        self.write('v%d' % var)

    def visit_param(self, param):
        self.write('p%s' % param)

    def visit_this(self):
        self.write('this')

    def visit_assign(self, lhs, rhs):
        self.write_ind()
        if lhs is None:
            rhs.visit(self)
            if not self.skip:
                self.end_ins()
            return
        lhs.visit(self)
        self.write(' = ')
        rhs.visit(self)
        self.end_ins()

    def visit_move_result(self, lhs, rhs):
        self.write_ind()
        lhs.visit(self)
        self.write(' = ')
        rhs.visit(self)
        self.end_ins()

    def visit_move(self, lhs, rhs):
        if lhs is rhs:
            return
        self.write_ind()
        lhs.visit(self)
        self.write(' = ')
        rhs.visit(self)
        self.end_ins()

    def visit_astore(self, array, index, rhs):
        self.write_ind()
        array.visit(self)
        self.write('[')
        if isinstance(index, Constant):
            index.visit(self, 'I')
        else:
            index.visit(self)
        self.write('] = ')
        rhs.visit(self)
        self.end_ins()

    def visit_put_static(self, cls, name, rhs):
        self.write_ind()
        self.write('%s.%s = ' % (cls, name))
        rhs.visit(self)
        self.end_ins()

    def visit_put_instance(self, lhs, name, rhs):
        self.write_ind()
        lhs.visit(self)
        self.write('.%s = ' % name)
        rhs.visit(self)
        self.end_ins()

    def visit_new(self, atype):
        self.write('new %s' % get_type(atype))

    def visit_invoke(self, name, base, args):
        if isinstance(base, ThisParam) and name == '<init>'\
            and self.constructor and len(args) == 0:
                self.skip = True
                return
        base.visit(self)
        if name != '<init>':
            self.write('.%s' % name)
        self.write('(')
        comma = False
        for arg in args:
            if comma:
                self.write(', ')
            comma = True
            arg.visit(self)
        self.write(')')

    def visit_return_void(self):
        self.write_ind()
        self.write('return')
        self.end_ins()

    def visit_return(self, arg):
        self.write_ind()
        self.write('return ')
        arg.visit(self)
        self.end_ins()

    def visit_nop(self):
        pass

    def visit_switch(self, arg):
        arg.visit(self)

    def visit_check_cast(self, arg, atype):
        self.write('(checkcast)(')
        arg.visit(self)
        self.write(', %s)' % atype)

    def visit_aload(self, array, index):
        array.visit(self)
        self.write('[')
        index.visit(self)
        self.write(']')

    def visit_alength(self, array):
        array.visit(self)
        self.write('.length')

    def visit_new_array(self, atype, size):
        self.write('new %s[' % get_type(atype[1:]))
        size.visit(self)
        self.write(']')

    def visit_filled_new_array(self, atype, size, args):
        self.write('filled-new-array(type=')
        atype.visit(self)
        self.write(', size=')
        size.visit(self)
        for arg in args:
            self.write(', arg=')
            arg.visit(self)
        self.write(')')

    def visit_fill_array(self, array, value):
        self.write_ind()
        array.visit(self)
        self.write(' = {')
        data = value.get_data()
        self.write(', '.join(['%d' % ord(c) for c in data[:-1]]))
        self.write('}')
        self.end_ins()

    def visit_monitor_enter(self, ref):
        self.write_ind()
        self.write('synchronized(')
        ref.visit(self)
        self.write(') {\n')
        self.inc_ind()

    def visit_monitor_exit(self, ref):
        self.dec_ind()
        self.write_ind()
        self.write('}\n')

    def visit_throw(self, ref):
        self.write_ind()
        self.write('throw ')
        ref.visit(self)
        self.end_ins()

    def visit_binary_expression(self, op, arg1, arg2):
        self.write('(')
        arg1.visit(self)
        self.write(' %s ' % op)
        arg2.visit(self)
        self.write(')')

    def visit_unary_expression(self, op, arg):
        self.write('(%s ' % op)
        arg.visit(self)
        self.write(')')

    def visit_cast(self, op, arg):
        self.write('(%s ' % op)
        arg.visit(self)
        self.write(')')

    def visit_cond_expression(self, op, arg1, arg2):
        arg1.visit(self)
        self.write(' %s ' % op)
        arg2.visit(self)

    def visit_condz_expression(self, op, arg):
        if isinstance(arg, BinaryCompExpression):
            arg.op = op
            return arg.visit(self)
        atype = arg.get_type()
        if atype == 'Z':
            if op is Op.EQUAL:
                self.write('!')
                arg.visit(self)
            else:
                arg.visit(self)
        else:
            arg.visit(self)
            self.write(' %s 0' % op)

    def visit_get_instance(self, arg, name):
        arg.visit(self)
        self.write('.%s' % name)

    def visit_get_static(self, cls, name):
        self.write('%s.%s' % (cls, name))


def string(s):
    # Based on http://stackoverflow.com/a/1676407
    ret = ['"']
    for c in s[1:-1]:
        if ord(c) < 32 or 0x80 <= ord(c) <= 0xff:
            to_add = '\\x%02x' % ord(c)
        elif c in '\\"':
            to_add = '%c' % c
        else:
            to_add = c
        ret.append(to_add)
    ret.append('"')
    return ''.join(ret)



# This file is part of Androguard.
#
# Copyright (c) 2012 Geoffroy Gueguen <geoffroy.gueguen@gmail.com>
# All Rights Reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

from struct import pack, unpack
from dad import util
from dad.instruction import (ArrayLengthExpression, ArrayLoadExpression,
                             ArrayStoreInstruction, AssignExpression,
                             BaseClass, BinaryCompExpression, BinaryExpression,
                             BinaryExpression2Addr, BinaryExpressionLit,
                             CastExpression, CheckCastExpression,
                             ConditionalExpression, ConditionalZExpression,
                             Constant, FillArrayExpression,
                             FilledArrayExpression, InstanceExpression,
                             InstanceInstruction, InvokeInstruction,
                             InvokeDirectInstruction, InvokeRangeInstruction,
                             InvokeStaticInstruction, MonitorEnterExpression,
                             MonitorExitExpression, MoveExpression,
                             MoveResultExpression, NewArrayExpression,
                             NewInstance, NopExpression, RefExpression,
                             ThrowExpression, Variable, ReturnInstruction,
                             StaticExpression, StaticInstruction,
                             SwitchExpression, UnaryExpression)

EXPR = 0
INST = 1
COND = 2


class Op(object):
    CMP = 'cmp'
    ADD = '+'
    SUB = '-'
    MUL = '*'
    DIV = '/'
    MOD = '%'
    AND = '&'
    OR = '|'
    XOR = '^'
    EQUAL = '=='
    NEQUAL = '!='
    GREATER = '>'
    LOWER = '<'
    GEQUAL = '>='
    LEQUAL = '<='
    NEG = '-'
    NOT = '~'
    INTSHL = '<<'  # '(%s << ( %s & 0x1f ))'
    INTSHR = '>>'  # '(%s >> ( %s & 0x1f ))'
    LONGSHL = '<<'  # '(%s << ( %s & 0x3f ))'
    LONGSHR = '>>'  # '(%s >> ( %s & 0x3f ))'


def get_variables(vmap, *variables):
    res = []
    for variable in variables:
        res.append(vmap.setdefault(variable, Variable(variable)))
    if len(res) == 1:
        return res[0]
    return res


# nop
def nop(ins, vmap):
    return NopExpression()


# move vA, vB ( 4b, 4b )
def move(ins, vmap):
    util.log('Move %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    return MoveExpression(a, b)


# move/from16 vAA, vBBBB ( 8b, 16b )
def movefrom16(ins, vmap):
    util.log('MoveFrom16 %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.AA, ins.BBBB)
    return MoveExpression(a, b)


# move/16 vAAAA, vBBBB ( 16b, 16b )
def move16(ins, vmap):
    util.log('Move16 %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.AAAA, ins.BBBB)
    return MoveExpression(a, b)


# move-wide vA, vB ( 4b, 4b )
def movewide(ins, vmap):
    util.log('MoveWide %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    return MoveExpression(a, b)


# move-wide/from16 vAA, vBBBB ( 8b, 16b )
def movewidefrom16(ins, vmap):
    util.log('MoveWideFrom16 : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.AA, ins.BBBB)
    return MoveExpression(a, b)


# move-wide/16 vAAAA, vBBBB ( 16b, 16b )
def movewide16(ins, vmap):
    util.log('MoveWide16 %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.AAAA, ins.BBBB)
    return MoveExpression(a, b)


# move-object vA, vB ( 4b, 4b )
def moveobject(ins, vmap):
    util.log('MoveObject %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    return MoveExpression(a, b)


# move-object/from16 vAA, vBBBB ( 8b, 16b )
def moveobjectfrom16(ins, vmap):
    util.log('MoveObjectFrom16 : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.AA, ins.BBBB)
    return MoveExpression(a, b)


# move-object/16 vAAAA, vBBBB ( 16b, 16b )
def moveobject16(ins, vmap):
    util.log('MoveObject16 : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.AAAA, ins.BBBB)
    return MoveExpression(a, b)


# move-result vAA ( 8b )
def moveresult(ins, vmap, ret):
    util.log('MoveResult : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return MoveResultExpression(a, ret)


# move-result-wide vAA ( 8b )
def moveresultwide(ins, vmap, ret):
    util.log('MoveResultWide : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return MoveResultExpression(a, ret)


# move-result-object vAA ( 8b )
def moveresultobject(ins, vmap, ret):
    util.log('MoveResultObject : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return MoveResultExpression(a, ret)


# move-exception vAA ( 8b )
def moveexception(ins, vmap):
    util.log('MoveException : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return RefExpression(a)


# return-void
def returnvoid(ins, vmap):
    util.log('ReturnVoid', 'debug')
    return ReturnInstruction(None)


# return vAA ( 8b )
def return_reg(ins, vmap):
    util.log('Return : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return ReturnInstruction(a)


# return-wide vAA ( 8b )
def returnwide(ins, vmap):
    util.log('ReturnWide : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return ReturnInstruction(a)


# return-object vAA ( 8b )
def returnobject(ins, vmap):
    util.log('ReturnObject : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return ReturnInstruction(a)


# const/4 vA, #+B ( 4b, 4b )
def const4(ins, vmap):
    util.log('Const4 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.B, 'I')
    a = get_variables(vmap, ins.A)
    exp = AssignExpression(a, cst)
    return exp


# const/16 vAA, #+BBBB ( 8b, 16b )
def const16(ins, vmap):
    util.log('Const16 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.BBBB, 'I')
    a = get_variables(vmap, ins.AA)
    exp = AssignExpression(a, cst)
    return exp


# const vAA, #+BBBBBBBB ( 8b, 32b )
def const(ins, vmap):
    util.log('Const : %s' % ins.get_output(), 'debug')
    value = unpack("=f", pack("=i", ins.BBBBBBBB))[0]
    cst = Constant(value, 'F', ins.BBBBBBBB)
    a = get_variables(vmap, ins.AA)
    exp = AssignExpression(a, cst)
    return exp


# const/high16 vAA, #+BBBB0000 ( 8b, 16b )
def consthigh16(ins, vmap):
    util.log('ConstHigh16 : %s' % ins.get_output(), 'debug')
    value = unpack('=f', '\x00\x00' + pack('=h', ins.BBBB))[0]
    cst = Constant(value, 'F', ins.BBBB)
    a = get_variables(vmap, ins.AA)
    exp = AssignExpression(a, cst)
    return exp


# const-wide/16 vAA, #+BBBB ( 8b, 16b )
def constwide16(ins, vmap):
    util.log('ConstWide16 : %s' % ins.get_output(), 'debug')
    value = unpack('=d', pack('=d', ins.BBBB))[0]
    cst = Constant(value, 'J', ins.BBBB)
    a = get_variables(vmap, ins.AA)
    exp = AssignExpression(a, cst)
    return exp


# const-wide/32 vAA, #+BBBBBBBB ( 8b, 32b )
def constwide32(ins, vmap):
    util.log('ConstWide32 : %s' % ins.get_output(), 'debug')
    value = unpack('=d', pack('=d', ins.BBBBBBBB))[0]
    cst = Constant(value, 'J', ins.BBBBBBBB)
    a = get_variables(vmap, ins.AA)
    exp = AssignExpression(a, cst)
    return exp


# const-wide vAA, #+BBBBBBBBBBBBBBBB ( 8b, 64b )
def constwide(ins, vmap):
    util.log('ConstWide : %s' % ins.get_output(), 'debug')
    value = unpack('=d', pack('=q', ins.BBBBBBBBBBBBBBBB))[0]
    cst = Constant(value, 'D', ins.BBBBBBBBBBBBBBBB)
    a = get_variables(vmap, ins.AA)
    exp = AssignExpression(a, cst)
    return exp


# const-wide/high16 vAA, #+BBBB000000000000 ( 8b, 16b )
def constwidehigh16(ins, vmap):
    util.log('ConstWideHigh16 : %s' % ins.get_output(), 'debug')
    value = unpack('=d',
                    '\x00\x00\x00\x00\x00\x00' + pack('=h', ins.BBBB))[0]
    cst = Constant(value, 'D', ins.BBBB)
    a = get_variables(vmap, ins.AA)
    exp = AssignExpression(a, cst)
    return exp


# const-string vAA ( 8b )
def conststring(ins, vmap):
    util.log('ConstString : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.get_string(), 'STR')
    a = get_variables(vmap, ins.AA)
    exp = AssignExpression(a, cst)
    return exp


# const-string/jumbo vAA ( 8b )
def conststringjumbo(ins, vmap):
    util.log('ConstStringJumbo %s' % ins.get_output(), 'debug')
    cst = Constant(ins.get_string(), 'STR')
    a = get_variables(vmap, ins.AA)
    exp = AssignExpression(a, cst)
    return exp


# const-class vAA, type@BBBB ( 8b )
def constclass(ins, vmap):
    util.log('ConstClass : %s' % ins.get_output(), 'debug')
    cst = Constant(util.get_type(ins.get_string()), 'class')
    a = get_variables(vmap, ins.AA)
    exp = AssignExpression(a, cst)
    return exp


# monitor-enter vAA ( 8b )
def monitorenter(ins, vmap):
    util.log('MonitorEnter : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return MonitorEnterExpression(a)


# monitor-exit vAA ( 8b )
def monitorexit(ins, vmap):
    util.log('MonitorExit : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return MonitorExitExpression(a)


# check-cast vAA ( 8b )
def checkcast(ins, vmap):
    util.log('CheckCast: %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    cast_type = util.get_type(ins.get_translated_kind())
    exp = CheckCastExpression(a, cast_type)
    return exp


# instance-of vA, vB ( 4b, 4b )
def instanceof(ins, vmap):
    util.log('InstanceOf : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    c = BaseClass(util.get_type(ins.get_translated_kind()))
    exp = BinaryExpression('instanceof', b, c)
    return AssignExpression(a, exp)


# array-length vA, vB ( 4b, 4b )
def arraylength(ins, vmap):
    util.log('ArrayLength: %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = ArrayLengthExpression(b)
    return AssignExpression(a, exp)


# new-instance vAA ( 8b )
def newinstance(ins, vmap):
    util.log('NewInstance : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    ins_type = ins.cm.get_type(ins.BBBB)
    exp = NewInstance(ins_type)
    return AssignExpression(a, exp)


# new-array vA, vB ( 8b, size )
def newarray(ins, vmap):
    util.log('NewArray : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = NewArrayExpression(b, ins.cm.get_type(ins.CCCC))
    return AssignExpression(a, exp)


# filled-new-array {vD, vE, vF, vG, vA} ( 4b each )
def fillednewarray(ins, vmap, ret):
    util.log('FilledNewArray : %s' % ins.get_output(), 'debug')
    a, b, c, d, e, f, g = get_variables(vmap, ins.A, ins.BBBB, ins.C, ins.D,
                                                    ins.E, ins.F, ins.G)
    exp = FilledArrayExpression(a, c, [d, e, f, g, a])
    return AssignExpression(ret, exp)


# filled-new-array/range {vCCCC..vNNNN} ( 16b )
def fillednewarrayrange(ins, vmap, ret):
    util.log('FilledNewArrayRange : %s' % ins.get_output(), 'debug')
    a, b, c, n = get_variables(vmap, ins.AA, ins.BBBB, ins.CCCC, ins.NNNN)
    exp = FilledArrayExpression(a, b, [c, n])
    return AssignExpression(ret, exp)


# fill-array-data vAA, +BBBBBBBB ( 8b, 32b )
def fillarraydata(ins, vmap, value):
    util.log('FillArrayData : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return FillArrayExpression(a, value)


# fill-array-data-payload vAA, +BBBBBBBB ( 8b, 32b )
def fillarraydatapayload(ins, vmap):
    util.log('FillArrayDataPayload : %s' % ins.get_output(), 'debug')
    return FillArrayExpression(None)


# throw vAA ( 8b )
def throw(ins, vmap):
    util.log('Throw : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return ThrowExpression(a)


# goto +AA ( 8b )
def goto(ins, vmap):
    return NopExpression()


# goto/16 +AAAA ( 16b )
def goto16(ins, vmap):
    return NopExpression()


# goto/32 +AAAAAAAA ( 32b )
def goto32(ins, vmap):
    return NopExpression()


# packed-switch vAA, +BBBBBBBB ( reg to test, 32b )
def packedswitch(ins, vmap):
    util.log('PackedSwitch : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.AA, ins.BBBBBBBB)
    return SwitchExpression(a, b)


# sparse-switch vAA, +BBBBBBBB ( reg to test, 32b )
def sparseswitch(ins, vmap):
    util.log('SparseSwitch : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.AA, ins.BBBBBBBB)
    return SwitchExpression(a, b)


# cmpl-float vAA, vBB, vCC ( 8b, 8b, 8b )
def cmplfloat(ins, vmap):
    util.log('CmpglFloat : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryCompExpression(Op.CMP, b, c)
    exp.type = 'F'
    return AssignExpression(a, exp)


# cmpg-float vAA, vBB, vCC ( 8b, 8b, 8b )
def cmpgfloat(ins, vmap):
    util.log('CmpgFloat : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryCompExpression(Op.CMP, b, c)
    exp.type = 'F'
    return AssignExpression(a, exp)


# cmpl-double vAA, vBB, vCC ( 8b, 8b, 8b )
def cmpldouble(ins, vmap):
    util.log('CmplDouble : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryCompExpression(Op.CMP, b, c)
    exp.type = 'D'
    return AssignExpression(a, exp)


# cmpg-double vAA, vBB, vCC ( 8b, 8b, 8b )
def cmpgdouble(ins, vmap):
    util.log('CmpgDouble : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryCompExpression(Op.CMP, b, c)
    exp.type = 'D'
    return AssignExpression(a, exp)


# cmp-long vAA, vBB, vCC ( 8b, 8b, 8b )
def cmplong(ins, vmap):
    util.log('CmpLong : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryCompExpression(Op.CMP, b, c)
    exp.type = 'J'
    return AssignExpression(a, exp)


# if-eq vA, vB, +CCCC ( 4b, 4b, 16b )
def ifeq(ins, vmap):
    util.log('IfEq : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    return ConditionalExpression(Op.EQUAL, a, b)


# if-ne vA, vB, +CCCC ( 4b, 4b, 16b )
def ifne(ins, vmap):
    util.log('IfNe : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    return ConditionalExpression(Op.NEQUAL, a, b)


# if-lt vA, vB, +CCCC ( 4b, 4b, 16b )
def iflt(ins, vmap):
    util.log('IfLt : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    return ConditionalExpression(Op.LOWER, a, b)


# if-ge vA, vB, +CCCC ( 4b, 4b, 16b )
def ifge(ins, vmap):
    util.log('IfGe : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    return ConditionalExpression(Op.GEQUAL, a, b)


# if-gt vA, vB, +CCCC ( 4b, 4b, 16b )
def ifgt(ins, vmap):
    util.log('IfGt : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    return ConditionalExpression(Op.GREATER, a, b)


# if-le vA, vB, +CCCC ( 4b, 4b, 16b )
def ifle(ins, vmap):
    util.log('IfLe : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    return ConditionalExpression(Op.LEQUAL, a, b)


# if-eqz vAA, +BBBB ( 8b, 16b )
def ifeqz(ins, vmap):
    util.log('IfEqz : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return ConditionalZExpression(Op.EQUAL, a)


# if-nez vAA, +BBBB ( 8b, 16b )
def ifnez(ins, vmap):
    util.log('IfNez : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return ConditionalZExpression(Op.NEQUAL, a)


# if-ltz vAA, +BBBB ( 8b, 16b )
def ifltz(ins, vmap):
    util.log('IfLtz : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return ConditionalZExpression(Op.LOWER, a)


# if-gez vAA, +BBBB ( 8b, 16b )
def ifgez(ins, vmap):
    util.log('IfGez : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return ConditionalZExpression(Op.GEQUAL, a)


# if-gtz vAA, +BBBB ( 8b, 16b )
def ifgtz(ins, vmap):
    util.log('IfGtz : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return ConditionalZExpression(Op.GREATER, a)


# if-lez vAA, +BBBB (8b, 16b )
def iflez(ins, vmap):
    util.log('IfLez : %s' % ins.get_output(), 'debug')
    a = get_variables(vmap, ins.AA)
    return ConditionalZExpression(Op.LEQUAL, a)


#TODO: check type for all aget
# aget vAA, vBB, vCC ( 8b, 8b, 8b )
def aget(ins, vmap):
    util.log('AGet : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = ArrayLoadExpression(b, c)
    return AssignExpression(a, exp)


# aget-wide vAA, vBB, vCC ( 8b, 8b, 8b )
def agetwide(ins, vmap):
    util.log('AGetWide : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = ArrayLoadExpression(b, c)
    return AssignExpression(a, exp)


# aget-object vAA, vBB, vCC ( 8b, 8b, 8b )
def agetobject(ins, vmap):
    util.log('AGetObject : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = ArrayLoadExpression(b, c)
    return AssignExpression(a, exp)


# aget-boolean vAA, vBB, vCC ( 8b, 8b, 8b )
def agetboolean(ins, vmap):
    util.log('AGetBoolean : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = ArrayLoadExpression(b, c)
    return AssignExpression(a, exp)


# aget-byte vAA, vBB, vCC ( 8b, 8b, 8b )
def agetbyte(ins, vmap):
    util.log('AGetByte : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = ArrayLoadExpression(b, c)
    return AssignExpression(a, exp)


# aget-char vAA, vBB, vCC ( 8b, 8b, 8b )
def agetchar(ins, vmap):
    util.log('AGetChar : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = ArrayLoadExpression(b, c)
    return AssignExpression(a, exp)


# aget-short vAA, vBB, vCC ( 8b, 8b, 8b )
def agetshort(ins, vmap):
    util.log('AGetShort : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = ArrayLoadExpression(b, c)
    return AssignExpression(a, exp)


# aput vAA, vBB, vCC
def aput(ins, vmap):
    util.log('APut : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    return ArrayStoreInstruction(a, b, c)


# aput-wide vAA, vBB, vCC ( 8b, 8b, 8b )
def aputwide(ins, vmap):
    util.log('APutWide : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    return ArrayStoreInstruction(a, b, c)


# aput-object vAA, vBB, vCC ( 8b, 8b, 8b )
def aputobject(ins, vmap):
    util.log('APutObject : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    return ArrayStoreInstruction(a, b, c)


# aput-boolean vAA, vBB, vCC ( 8b, 8b, 8b )
def aputboolean(ins, vmap):
    util.log('APutBoolean : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    return ArrayStoreInstruction(a, b, c)


# aput-byte vAA, vBB, vCC ( 8b, 8b, 8b )
def aputbyte(ins, vmap):
    util.log('APutByte : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    return ArrayStoreInstruction(a, b, c)


# aput-char vAA, vBB, vCC ( 8b, 8b, 8b )
def aputchar(ins, vmap):
    util.log('APutChar : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    return ArrayStoreInstruction(a, b, c)


# aput-short vAA, vBB, vCC ( 8b, 8b, 8b )
def aputshort(ins, vmap):
    util.log('APutShort : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    return ArrayStoreInstruction(a, b, c)


# iget vA, vB ( 4b, 4b )
def iget(ins, vmap):
    util.log('IGet : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = InstanceExpression(b, klass, ftype, name)
    return AssignExpression(a, exp)


# iget-wide vA, vB ( 4b, 4b )
def igetwide(ins, vmap):
    util.log('IGetWide : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = InstanceExpression(b, klass, ftype, name)
    return AssignExpression(a, exp)


# iget-object vA, vB ( 4b, 4b )
def igetobject(ins, vmap):
    util.log('IGetObject : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = InstanceExpression(b, klass, ftype, name)
    return AssignExpression(a, exp)


# iget-boolean vA, vB ( 4b, 4b )
def igetboolean(ins, vmap):
    util.log('IGetBoolean : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = InstanceExpression(b, klass, ftype, name)
    return AssignExpression(a, exp)


# iget-byte vA, vB ( 4b, 4b )
def igetbyte(ins, vmap):
    util.log('IGetByte : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = InstanceExpression(b, klass, ftype, name)
    return AssignExpression(a, exp)


# iget-char vA, vB ( 4b, 4b )
def igetchar(ins, vmap):
    util.log('IGetChar : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = InstanceExpression(b, klass, ftype, name)
    return AssignExpression(a, exp)


# iget-short vA, vB ( 4b, 4b )
def igetshort(ins, vmap):
    util.log('IGetShort : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = InstanceExpression(b, klass, ftype, name)
    return AssignExpression(a, exp)


# iput vA, vB ( 4b, 4b )
def iput(ins, vmap):
    util.log('IPut %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    return InstanceInstruction(a, b, klass, atype, name)


# iput-wide vA, vB ( 4b, 4b )
def iputwide(ins, vmap):
    util.log('IPutWide %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    return InstanceInstruction(a, b, klass, atype, name)


# iput-object vA, vB ( 4b, 4b )
def iputobject(ins, vmap):
    util.log('IPutObject %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    return InstanceInstruction(a, b, klass, atype, name)


# iput-boolean vA, vB ( 4b, 4b )
def iputboolean(ins, vmap):
    util.log('IPutBoolean %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    return InstanceInstruction(a, b, klass, atype, name)


# iput-byte vA, vB ( 4b, 4b )
def iputbyte(ins, vmap):
    util.log('IPutByte %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    return InstanceInstruction(a, b, klass, atype, name)


# iput-char vA, vB ( 4b, 4b )
def iputchar(ins, vmap):
    util.log('IPutChar %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    return InstanceInstruction(a, b, klass, atype, name)


# iput-short vA, vB ( 4b, 4b )
def iputshort(ins, vmap):
    util.log('IPutShort %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.CCCC)
    klass = util.get_type(klass)
    a, b = get_variables(vmap, ins.A, ins.B)
    return InstanceInstruction(a, b, klass, atype, name)


# sget vAA ( 8b )
def sget(ins, vmap):
    util.log('SGet : %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    exp = StaticExpression(klass, atype, name)
    a = get_variables(vmap, ins.AA)
    return AssignExpression(a, exp)


# sget-wide vAA ( 8b )
def sgetwide(ins, vmap):
    util.log('SGetWide : %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    exp = StaticExpression(klass, atype, name)
    a = get_variables(vmap, ins.AA)
    return AssignExpression(a, exp)


# sget-object vAA ( 8b )
def sgetobject(ins, vmap):
    util.log('SGetObject : %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    exp = StaticExpression(klass, atype, name)
    a = get_variables(vmap, ins.AA)
    return AssignExpression(a, exp)


# sget-boolean vAA ( 8b )
def sgetboolean(ins, vmap):
    util.log('SGetBoolean : %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    exp = StaticExpression(klass, atype, name)
    a = get_variables(vmap, ins.AA)
    return AssignExpression(a, exp)


# sget-byte vAA ( 8b )
def sgetbyte(ins, vmap):
    util.log('SGetByte : %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    exp = StaticExpression(klass, atype, name)
    a = get_variables(vmap, ins.AA)
    return AssignExpression(a, exp)


# sget-char vAA ( 8b )
def sgetchar(ins, vmap):
    util.log('SGetChar : %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    exp = StaticExpression(klass, atype, name)
    a = get_variables(vmap, ins.AA)
    return AssignExpression(a, exp)


# sget-short vAA ( 8b )
def sgetshort(ins, vmap):
    util.log('SGetShort : %s' % ins.get_output(), 'debug')
    klass, atype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    exp = StaticExpression(klass, atype, name)
    a = get_variables(vmap, ins.AA)
    return AssignExpression(a, exp)


# sput vAA ( 8b )
def sput(ins, vmap):
    util.log('SPut : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    a = get_variables(vmap, ins.AA)
    return StaticInstruction(a, klass, ftype, name)


# sput-wide vAA ( 8b )
def sputwide(ins, vmap):
    util.log('SPutWide : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    a = get_variables(vmap, ins.AA)
    return StaticInstruction(a, klass, ftype, name)


# sput-object vAA ( 8b )
def sputobject(ins, vmap):
    util.log('SPutObject : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    a = get_variables(vmap, ins.AA)
    return StaticInstruction(a, klass, ftype, name)


# sput-boolean vAA ( 8b )
def sputboolean(ins, vmap):
    util.log('SPutBoolean : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    a = get_variables(vmap, ins.AA)
    return StaticInstruction(a, klass, ftype, name)


# sput-wide vAA ( 8b )
def sputbyte(ins, vmap):
    util.log('SPutByte : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    a = get_variables(vmap, ins.AA)
    return StaticInstruction(a, klass, ftype, name)


# sput-char vAA ( 8b )
def sputchar(ins, vmap):
    util.log('SPutChar : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    a = get_variables(vmap, ins.AA)
    return StaticInstruction(a, klass, ftype, name)


# sput-short vAA ( 8b )
def sputshort(ins, vmap):
    util.log('SPutShort : %s' % ins.get_output(), 'debug')
    klass, ftype, name = ins.cm.get_field(ins.BBBB)
    klass = util.get_type(klass)
    a = get_variables(vmap, ins.AA)
    return StaticInstruction(a, klass, ftype, name)


# invoke-virtual {vD, vE, vF, vG, vA} ( 4b each )
def invokevirtual(ins, vmap, ret):
    util.log('InvokeVirtual : %s' % ins.get_output(), 'debug')
    method = ins.cm.get_method_ref(ins.BBBB)
    cls_name = util.get_type(method.get_class_name())
    name = method.get_name()
    param_type, ret_type = method.get_proto()
    ret_type = util.get_type(ret_type)
    param_type = util.get_params_type(param_type)
    nbargs = ins.A - 1
    largs = [ins.D, ins.E, ins.F, ins.G]
    args = get_variables(vmap, *largs)[:nbargs]
    c = get_variables(vmap, ins.C)
    exp = InvokeInstruction(cls_name, name, c, ret_type,
                            param_type, nbargs, args)
    return AssignExpression(ret.new(), exp)


# invoke-super {vD, vE, vF, vG, vA} ( 4b each )
def invokesuper(ins, vmap, ret):
    util.log('InvokeSuper : %s' % ins.get_output(), 'debug')
    method = ins.cm.get_method_ref(ins.BBBB)
    cls_name = util.get_type(method.get_class_name())
    name = method.get_name()
    param_type, ret_type = method.get_proto()
    ret_type = util.get_type(ret_type)
    param_type = util.get_params_type(param_type)
    nbargs = ins.A - 1
    largs = [ins.D, ins.E, ins.F, ins.G]
    args = get_variables(vmap, *largs)[:nbargs]
    superclass = BaseClass('super')
    exp = InvokeInstruction(cls_name, name, superclass, ret_type,
                            param_type, nbargs, args)
    return AssignExpression(ret.new(), exp)


# invoke-direct {vD, vE, vF, vG, vA} ( 4b each )
def invokedirect(ins, vmap, ret):
    util.log('InvokeDirect : %s' % ins.get_output(), 'debug')
    method = ins.cm.get_method_ref(ins.BBBB)
    cls_name = util.get_type(method.get_class_name())
    name = method.get_name()
    param_type, ret_type = method.get_proto()
    ret_type = util.get_type(ret_type)
    param_type = util.get_params_type(param_type)
#    nbargs = ins.A
    nbargs = ins.A - 1
#    largs = [ins.C, ins.D, ins.E, ins.F, ins.G]
    largs = [ins.D, ins.E, ins.F, ins.G]
    args = get_variables(vmap, *largs)[:nbargs]
    c = get_variables(vmap, ins.C)
    ret.set_to(c)
#    exp = InvokeDirectInstruction(cls_name, name, ret_type,
#                                    param_type, nbargs, args)
#    return AssignExpression(c, exp)
    exp = InvokeDirectInstruction(cls_name, name, c, ret_type,
                            param_type, nbargs, args)
    return AssignExpression(c, exp)


# invoke-static {vD, vE, vF, vG, vA} ( 4b each )
def invokestatic(ins, vmap, ret):
    util.log('InvokeStatic : %s' % ins.get_output(), 'debug')
    method = ins.cm.get_method_ref(ins.BBBB)
    cls_name = util.get_type(method.get_class_name())
    name = method.get_name()
    param_type, ret_type = method.get_proto()
    ret_type = util.get_type(ret_type)
    param_type = util.get_params_type(param_type)
    nbargs = ins.A
    largs = [ins.C, ins.D, ins.E, ins.F, ins.G]
    args = get_variables(vmap, *largs)[:nbargs]
    base = BaseClass(util.get_type(cls_name))
    exp = InvokeStaticInstruction(cls_name, name, base, ret_type,
                                    param_type, nbargs, args)
    return AssignExpression(ret.new(), exp)


# invoke-interface {vD, vE, vF, vG, vA} ( 4b each )
def invokeinterface(ins, vmap, ret):
    util.log('InvokeInterface : %s' % ins.get_output(), 'debug')
    method = ins.cm.get_method_ref(ins.BBBB)
    cls_name = util.get_type(method.get_class_name())
    name = method.get_name()
    param_type, ret_type = method.get_proto()
    ret_type = util.get_type(ret_type)
    param_type = util.get_params_type(param_type)
    nbargs = ins.A - 1
    largs = [ins.D, ins.E, ins.F, ins.G]
    args = get_variables(vmap, *largs)[:nbargs]
    c = get_variables(vmap, ins.C)
    exp = InvokeInstruction(cls_name, name, c, ret_type,
                            param_type, nbargs, args)
    return AssignExpression(ret.new(), exp)


# invoke-virtual/range {vCCCC..vNNNN} ( 16b each )
def invokevirtualrange(ins, vmap, ret):
    util.log('InvokeVirtualRange : %s' % ins.get_output(), 'debug')
    method = ins.cm.get_method_ref(ins.BBBB)
    cls_name = util.get_type(method.get_class_name())
    name = method.get_name()
    param_type, ret_type = method.get_proto()
    ret_type = util.get_type(ret_type)
    param_type = util.get_params_type(param_type)
    nbargs = ins.AA
    largs = range(ins.CCCC, ins.NNNN + 1)
    args = get_variables(vmap, *largs)
    if len(largs) == 1:
        args = [args]
    exp = InvokeRangeInstruction(cls_name, name, ret_type,
                                 param_type, nbargs, args)
    return AssignExpression(ret.new(), exp)


# invoke-super/range {vCCCC..vNNNN} ( 16b each )
def invokesuperrange(ins, vmap, ret):
    util.log('InvokeSuperRange : %s' % ins.get_output(), 'debug')
    method = ins.cm.get_method_ref(ins.BBBB)
    cls_name = util.get_type(method.get_class_name())
    name = method.get_name()
    param_type, ret_type = method.get_proto()
    ret_type = util.get_type(ret_type)
    param_type = util.get_params_type(param_type)
    nbargs = ins.AA
    largs = range(ins.CCCC, ins.NNNN + 1)
    args = get_variables(vmap, *largs)
    if len(largs) == 1:
        args = [args]
    exp = InvokeRangeInstruction(cls_name, name, ret_type,
                                param_type, nbargs, args)
    return AssignExpression(ret.new(), exp)


# invoke-direct/range {vCCCC..vNNNN} ( 16b each )
def invokedirectrange(ins, vmap, ret):
    util.log('InvokeDirectRange : %s' % ins.get_output(), 'debug')
    method = ins.cm.get_method_ref(ins.BBBB)
    cls_name = util.get_type(method.get_class_name())
    name = method.get_name()
    param_type, ret_type = method.get_proto()
    ret_type = util.get_type(ret_type)
    param_type = util.get_params_type(param_type)
    nbargs = ins.AA
    largs = range(ins.CCCC, ins.NNNN + 1)
    args = get_variables(vmap, *largs)
    if len(largs) == 1:
        args = [args]
    c = get_variables(vmap, ins.CCCC)
    ret.set_to(c)
    exp = InvokeRangeInstruction(cls_name, name, ret_type,
                                param_type, nbargs, args)
    return AssignExpression(c, exp)


# invoke-static/range {vCCCC..vNNNN} ( 16b each )
def invokestaticrange(ins, vmap, ret):
    util.log('InvokeStaticRange : %s' % ins.get_output(), 'debug')
    method = ins.cm.get_method_ref(ins.BBBB)
    cls_name = util.get_type(method.get_class_name())
    name = method.get_name()
    param_type, ret_type = method.get_proto()
    ret_type = util.get_type(ret_type)
    param_type = util.get_params_type(param_type)
    nbargs = ins.AA
    largs = range(ins.CCCC, ins.NNNN + 1)
    args = get_variables(vmap, *largs)
    if len(largs) == 1:
        args = [args]
    base = BaseClass(util.get_type(cls_name))
    exp = InvokeStaticInstruction(cls_name, name, base, ret_type,
                                param_type, nbargs, args)
    return AssignExpression(ret.new(), exp)


# invoke-interface/range {vCCCC..vNNNN} ( 16b each )
def invokeinterfacerange(ins, vmap, ret):
    util.log('InvokeInterfaceRange : %s' % ins.get_output(), 'debug')
    method = ins.cm.get_method_ref(ins.BBBB)
    cls_name = util.get_type(method.get_class_name())
    name = method.get_name()
    param_type, ret_type = method.get_proto()
    ret_type = util.get_type(ret_type)
    param_type = util.get_params_type(param_type)
    nbargs = ins.AA
    largs = range(ins.CCCC, ins.NNNN + 1)
    args = get_variables(vmap, *largs)
    if len(largs) == 1:
        args = [args]
    exp = InvokeRangeInstruction(cls_name, name, ret_type,
                                param_type, nbargs, args)
    return AssignExpression(ret.new(), exp)


# neg-int vA, vB ( 4b, 4b )
def negint(ins, vmap):
    util.log('NegInt : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = UnaryExpression(Op.NEG, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# not-int vA, vB ( 4b, 4b )
def notint(ins, vmap):
    util.log('NotInt : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = UnaryExpression(Op.NOT, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# neg-long vA, vB ( 4b, 4b )
def neglong(ins, vmap):
    util.log('NegLong : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = UnaryExpression(Op.NEG, b)
    exp.type = 'J'
    return AssignExpression(a, exp)


# not-long vA, vB ( 4b, 4b )
def notlong(ins, vmap):
    util.log('NotLong : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = UnaryExpression(Op.NOT, b)
    exp.type = 'J'
    return AssignExpression(a, exp)


# neg-float vA, vB ( 4b, 4b )
def negfloat(ins, vmap):
    util.log('NegFloat : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = UnaryExpression(Op.NEG, b)
    exp.type = 'F'
    return AssignExpression(a, exp)


# neg-double vA, vB ( 4b, 4b )
def negdouble(ins, vmap):
    util.log('NegDouble : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = UnaryExpression(Op.NEG, b)
    exp.type = 'D'
    return AssignExpression(a, exp)


# int-to-long vA, vB ( 4b, 4b )
def inttolong(ins, vmap):
    util.log('IntToLong : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(long)', 'J', b)
    return AssignExpression(a, exp)


# int-to-float vA, vB ( 4b, 4b )
def inttofloat(ins, vmap):
    util.log('IntToFloat : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(float)', 'F', b)
    return AssignExpression(a, exp)


# int-to-double vA, vB ( 4b, 4b )
def inttodouble(ins, vmap):
    util.log('IntToDouble : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(double)', 'D', b)
    return AssignExpression(a, exp)


# long-to-int vA, vB ( 4b, 4b )
def longtoint(ins, vmap):
    util.log('LongToInt : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(int)', 'I', b)
    return AssignExpression(a, exp)


# long-to-float vA, vB ( 4b, 4b )
def longtofloat(ins, vmap):
    util.log('LongToFloat : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(float)', 'F', b)
    return AssignExpression(a, exp)


# long-to-double vA, vB ( 4b, 4b )
def longtodouble(ins, vmap):
    util.log('LongToDouble : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(double)', 'D', b)
    return AssignExpression(a, exp)


# float-to-int vA, vB ( 4b, 4b )
def floattoint(ins, vmap):
    util.log('FloatToInt : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(int)', 'I', b)
    return AssignExpression(a, exp)


# float-to-long vA, vB ( 4b, 4b )
def floattolong(ins, vmap):
    util.log('FloatToLong : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(long)', 'J', b)
    return AssignExpression(a, exp)


# float-to-double vA, vB ( 4b, 4b )
def floattodouble(ins, vmap):
    util.log('FloatToDouble : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(double)', 'D', b)
    return AssignExpression(a, exp)


# double-to-int vA, vB ( 4b, 4b )
def doubletoint(ins, vmap):
    util.log('DoubleToInt : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(int)', 'I', b)
    return AssignExpression(a, exp)


# double-to-long vA, vB ( 4b, 4b )
def doubletolong(ins, vmap):
    util.log('DoubleToLong : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(long)', 'J', b)
    return AssignExpression(a, exp)


# double-to-float vA, vB ( 4b, 4b )
def doubletofloat(ins, vmap):
    util.log('DoubleToFloat : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(float)', 'F', b)
    return AssignExpression(a, exp)


# int-to-byte vA, vB ( 4b, 4b )
def inttobyte(ins, vmap):
    util.log('IntToByte : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(byte)', 'B', b)
    return AssignExpression(a, exp)


# int-to-char vA, vB ( 4b, 4b )
def inttochar(ins, vmap):
    util.log('IntToChar : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(char)', 'C', b)
    return AssignExpression(a, exp)


# int-to-short vA, vB ( 4b, 4b )
def inttoshort(ins, vmap):
    util.log('IntToShort : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = CastExpression('(short)', 'S', b)
    return AssignExpression(a, exp)


# add-int vAA, vBB, vCC ( 8b, 8b, 8b )
def addint(ins, vmap):
    util.log('AddInt : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.ADD, b, c)
    exp.type = 'I'
    return AssignExpression(a, exp)


# sub-int vAA, vBB, vCC ( 8b, 8b, 8b )
def subint(ins, vmap):
    util.log('SubInt : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.SUB, b, c)
    exp.type = 'I'
    return AssignExpression(a, exp)


# mul-int vAA, vBB, vCC ( 8b, 8b, 8b )
def mulint(ins, vmap):
    util.log('MulInt : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.MUL, b, c)
    exp.type = 'I'
    return AssignExpression(a, exp)


# div-int vAA, vBB, vCC ( 8b, 8b, 8b )
def divint(ins, vmap):
    util.log('DivInt : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.DIV, b, c)
    exp.type = 'I'
    return AssignExpression(a, exp)


# rem-int vAA, vBB, vCC ( 8b, 8b, 8b )
def remint(ins, vmap):
    util.log('RemInt : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.MOD, b, c)
    exp.type = 'I'
    return AssignExpression(a, exp)


# and-int vAA, vBB, vCC ( 8b, 8b, 8b )
def andint(ins, vmap):
    util.log('AndInt : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.AND, b, c)
    exp.type = 'I'
    return AssignExpression(a, exp)


# or-int vAA, vBB, vCC ( 8b, 8b, 8b )
def orint(ins, vmap):
    util.log('OrInt : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.OR, b, c)
    exp.type = 'I'
    return AssignExpression(a, exp)


# xor-int vAA, vBB, vCC ( 8b, 8b, 8b )
def xorint(ins, vmap):
    util.log('XorInt : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.XOR, b, c)
    exp.type = 'I'
    return AssignExpression(a, exp)


# shl-int vAA, vBB, vCC ( 8b, 8b, 8b )
def shlint(ins, vmap):
    util.log('ShlInt : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.INTSHL, b, c)
    exp.type = 'I'
    return AssignExpression(a, exp)


# shr-int vAA, vBB, vCC ( 8b, 8b, 8b )
def shrint(ins, vmap):
    util.log('ShrInt : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.INTSHR, b, c)
    exp.type = 'I'
    return AssignExpression(a, exp)


# ushr-int vAA, vBB, vCC ( 8b, 8b, 8b )
def ushrint(ins, vmap):
    util.log('UShrInt : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.INTSHR, b, c)
    exp.type = 'I'
    return AssignExpression(a, exp)


# add-long vAA, vBB, vCC ( 8b, 8b, 8b )
def addlong(ins, vmap):
    util.log('AddLong : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.ADD, b, c)
    exp.type = 'J'
    return AssignExpression(a, exp)


# sub-long vAA, vBB, vCC ( 8b, 8b, 8b )
def sublong(ins, vmap):
    util.log('SubLong : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.SUB, b, c)
    exp.type = 'J'
    return AssignExpression(a, exp)


# mul-long vAA, vBB, vCC ( 8b, 8b, 8b )
def mullong(ins, vmap):
    util.log('MulLong : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.MUL, b, c)
    exp.type = 'J'
    return AssignExpression(a, exp)


# div-long vAA, vBB, vCC ( 8b, 8b, 8b )
def divlong(ins, vmap):
    util.log('DivLong : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.DIV, b, c)
    exp.type = 'J'
    return AssignExpression(a, exp)


# rem-long vAA, vBB, vCC ( 8b, 8b, 8b )
def remlong(ins, vmap):
    util.log('RemLong : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.MOD, b, c)
    exp.type = 'J'
    return AssignExpression(a, exp)


# and-long vAA, vBB, vCC ( 8b, 8b, 8b )
def andlong(ins, vmap):
    util.log('AndLong : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.AND, b, c)
    exp.type = 'J'
    return AssignExpression(a, exp)


# or-long vAA, vBB, vCC ( 8b, 8b, 8b )
def orlong(ins, vmap):
    util.log('OrLong : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.OR, b, c)
    exp.type = 'J'
    return AssignExpression(a, exp)


# xor-long vAA, vBB, vCC ( 8b, 8b, 8b )
def xorlong(ins, vmap):
    util.log('XorLong : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.XOR, b, c)
    exp.type = 'J'
    return AssignExpression(a, exp)


# shl-long vAA, vBB, vCC ( 8b, 8b, 8b )
def shllong(ins, vmap):
    util.log('ShlLong : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.LONGSHL, b, c)
    exp.type = 'J'
    return AssignExpression(a, exp)


# shr-long vAA, vBB, vCC ( 8b, 8b, 8b )
def shrlong(ins, vmap):
    util.log('ShrLong : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.LONGSHR, b, c)
    exp.type = 'J'
    return AssignExpression(a, exp)


# ushr-long vAA, vBB, vCC ( 8b, 8b, 8b )
def ushrlong(ins, vmap):
    util.log('UShrLong : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.LONGSHR, b, c)
    exp.type = 'J'
    return AssignExpression(a, exp)


# add-float vAA, vBB, vCC ( 8b, 8b, 8b )
def addfloat(ins, vmap):
    util.log('AddFloat : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.ADD, b, c)
    exp.type = 'F'
    return AssignExpression(a, exp)


# sub-float vAA, vBB, vCC ( 8b, 8b, 8b )
def subfloat(ins, vmap):
    util.log('SubFloat : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.SUB, b, c)
    exp.type = 'F'
    return AssignExpression(a, exp)


# mul-float vAA, vBB, vCC ( 8b, 8b, 8b )
def mulfloat(ins, vmap):
    util.log('MulFloat : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.MUL, b, c)
    exp.type = 'F'
    return AssignExpression(a, exp)


# div-float vAA, vBB, vCC ( 8b, 8b, 8b )
def divfloat(ins, vmap):
    util.log('DivFloat : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.DIV, b, c)
    exp.type = 'F'
    return AssignExpression(a, exp)


# rem-float vAA, vBB, vCC ( 8b, 8b, 8b )
def remfloat(ins, vmap):
    util.log('RemFloat : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.MOD, b, c)
    exp.type = 'F'
    return AssignExpression(a, exp)


# add-double vAA, vBB, vCC ( 8b, 8b, 8b )
def adddouble(ins, vmap):
    util.log('AddDouble : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.ADD, b, c)
    exp.type = 'D'
    return AssignExpression(a, exp)


# sub-double vAA, vBB, vCC ( 8b, 8b, 8b )
def subdouble(ins, vmap):
    util.log('SubDouble : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.SUB, b, c)
    exp.type = 'D'
    return AssignExpression(a, exp)


# mul-double vAA, vBB, vCC ( 8b, 8b, 8b )
def muldouble(ins, vmap):
    util.log('MulDouble : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.MUL, b, c)
    exp.type = 'D'
    return AssignExpression(a, exp)


# div-double vAA, vBB, vCC ( 8b, 8b, 8b )
def divdouble(ins, vmap):
    util.log('DivDouble : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.DIV, b, c)
    exp.type = 'D'
    return AssignExpression(a, exp)


# rem-double vAA, vBB, vCC ( 8b, 8b, 8b )
def remdouble(ins, vmap):
    util.log('RemDouble : %s' % ins.get_output(), 'debug')
    a, b, c = get_variables(vmap, ins.AA, ins.BB, ins.CC)
    exp = BinaryExpression(Op.MOD, b, c)
    exp.type = 'D'
    return AssignExpression(a, exp)


# add-int/2addr vA, vB ( 4b, 4b )
def addint2addr(ins, vmap):
    util.log('AddInt2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.ADD, a, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# sub-int/2addr vA, vB ( 4b, 4b )
def subint2addr(ins, vmap):
    util.log('SubInt2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.SUB, a, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# mul-int/2addr vA, vB ( 4b, 4b )
def mulint2addr(ins, vmap):
    util.log('MulInt2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.MUL, a, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# div-int/2addr vA, vB ( 4b, 4b )
def divint2addr(ins, vmap):
    util.log('DivInt2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.DIV, a, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# rem-int/2addr vA, vB ( 4b, 4b )
def remint2addr(ins, vmap):
    util.log('RemInt2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.MOD, a, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# and-int/2addr vA, vB ( 4b, 4b )
def andint2addr(ins, vmap):
    util.log('AndInt2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.AND, a, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# or-int/2addr vA, vB ( 4b, 4b )
def orint2addr(ins, vmap):
    util.log('OrInt2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.OR, a, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# xor-int/2addr vA, vB ( 4b, 4b )
def xorint2addr(ins, vmap):
    util.log('XorInt2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.XOR, a, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# shl-int/2addr vA, vB ( 4b, 4b )
def shlint2addr(ins, vmap):
    util.log('ShlInt2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.INTSHL, a, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# shr-int/2addr vA, vB ( 4b, 4b )
def shrint2addr(ins, vmap):
    util.log('ShrInt2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.INTSHR, a, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# ushr-int/2addr vA, vB ( 4b, 4b )
def ushrint2addr(ins, vmap):
    util.log('UShrInt2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.INTSHR, a, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# add-long/2addr vA, vB ( 4b, 4b )
def addlong2addr(ins, vmap):
    util.log('AddLong2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.ADD, a, b)
    exp.type = 'J'
    return AssignExpression(a, exp)


# sub-long/2addr vA, vB ( 4b, 4b )
def sublong2addr(ins, vmap):
    util.log('SubLong2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.SUB, a, b)
    exp.type = 'J'
    return AssignExpression(a, exp)


# mul-long/2addr vA, vB ( 4b, 4b )
def mullong2addr(ins, vmap):
    util.log('MulLong2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.MUL, a, b)
    exp.type = 'J'
    return AssignExpression(a, exp)


# div-long/2addr vA, vB ( 4b, 4b )
def divlong2addr(ins, vmap):
    util.log('DivLong2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.DIV, a, b)
    exp.type = 'J'
    return AssignExpression(a, exp)


# rem-long/2addr vA, vB ( 4b, 4b )
def remlong2addr(ins, vmap):
    util.log('RemLong2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.MUL, a, b)
    exp.type = 'J'
    return AssignExpression(a, exp)


# and-long/2addr vA, vB ( 4b, 4b )
def andlong2addr(ins, vmap):
    util.log('AndLong2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.AND, a, b)
    exp.type = 'J'
    return AssignExpression(a, exp)


# or-long/2addr vA, vB ( 4b, 4b )
def orlong2addr(ins, vmap):
    util.log('OrLong2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.OR, a, b)
    exp.type = 'J'
    return AssignExpression(a, exp)


# xor-long/2addr vA, vB ( 4b, 4b )
def xorlong2addr(ins, vmap):
    util.log('XorLong2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.XOR, a, b)
    exp.type = 'J'
    return AssignExpression(a, exp)


# shl-long/2addr vA, vB ( 4b, 4b )
def shllong2addr(ins, vmap):
    util.log('ShlLong2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.LONGSHL, a, b)
    exp.type = 'J'
    return AssignExpression(a, exp)


# shr-long/2addr vA, vB ( 4b, 4b )
def shrlong2addr(ins, vmap):
    util.log('ShrLong2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.LONGSHR, a, b)
    exp.type = 'J'
    return AssignExpression(a, exp)


# ushr-long/2addr vA, vB ( 4b, 4b )
def ushrlong2addr(ins, vmap):
    util.log('UShrLong2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.LONGSHR, a, b)
    exp.type = 'J'
    return AssignExpression(a, exp)


# add-float/2addr vA, vB ( 4b, 4b )
def addfloat2addr(ins, vmap):
    util.log('AddFloat2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.ADD, a, b)
    exp.type = 'F'
    return AssignExpression(a, exp)


# sub-float/2addr vA, vB ( 4b, 4b )
def subfloat2addr(ins, vmap):
    util.log('SubFloat2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.SUB, a, b)
    exp.type = 'F'
    return AssignExpression(a, exp)


# mul-float/2addr vA, vB ( 4b, 4b )
def mulfloat2addr(ins, vmap):
    util.log('MulFloat2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.MUL, a, b)
    exp.type = 'F'
    return AssignExpression(a, exp)


# div-float/2addr vA, vB ( 4b, 4b )
def divfloat2addr(ins, vmap):
    util.log('DivFloat2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.DIV, a, b)
    exp.type = 'F'
    return AssignExpression(a, exp)


# rem-float/2addr vA, vB ( 4b, 4b )
def remfloat2addr(ins, vmap):
    util.log('RemFloat2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.MOD, a, b)
    exp.type = 'F'
    return AssignExpression(a, exp)


# add-double/2addr vA, vB ( 4b, 4b )
def adddouble2addr(ins, vmap):
    util.log('AddDouble2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.ADD, a, b)
    exp.type = 'D'
    return AssignExpression(a, exp)


# sub-double/2addr vA, vB ( 4b, 4b )
def subdouble2addr(ins, vmap):
    util.log('subDouble2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.SUB, a, b)
    exp.type = 'D'
    return AssignExpression(a, exp)


# mul-double/2addr vA, vB ( 4b, 4b )
def muldouble2addr(ins, vmap):
    util.log('MulDouble2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.MUL, a, b)
    exp.type = 'D'
    return AssignExpression(a, exp)


# div-double/2addr vA, vB ( 4b, 4b )
def divdouble2addr(ins, vmap):
    util.log('DivDouble2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.DIV, a, b)
    exp.type = 'D'
    return AssignExpression(a, exp)


# rem-double/2addr vA, vB ( 4b, 4b )
def remdouble2addr(ins, vmap):
    util.log('RemDouble2Addr : %s' % ins.get_output(), 'debug')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpression2Addr(Op.MOD, a, b)
    exp.type = 'D'
    return AssignExpression(a, exp)


# add-int/lit16 vA, vB, #+CCCC ( 4b, 4b, 16b )
def addintlit16(ins, vmap):
    util.log('AddIntLit16 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CCCC, 'I')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpressionLit(Op.ADD, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# rsub-int vA, vB, #+CCCC ( 4b, 4b, 16b )
def rsubint(ins, vmap):
    util.log('RSubInt : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CCCC, 'I')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpressionLit(Op.SUB, cst, b)
    exp.type = 'I'
    return AssignExpression(a, exp)


# mul-int/lit16 vA, vB, #+CCCC ( 4b, 4b, 16b )
def mulintlit16(ins, vmap):
    util.log('MulIntLit16 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CCCC, 'I')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpressionLit(Op.MUL, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# div-int/lit16 vA, vB, #+CCCC ( 4b, 4b, 16b )
def divintlit16(ins, vmap):
    util.log('DivIntLit16 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CCCC, 'I')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpressionLit(Op.DIV, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# rem-int/lit16 vA, vB, #+CCCC ( 4b, 4b, 16b )
def remintlit16(ins, vmap):
    util.log('RemIntLit16 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CCCC, 'I')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpressionLit(Op.MOD, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# and-int/lit16 vA, vB, #+CCCC ( 4b, 4b, 16b )
def andintlit16(ins, vmap):
    util.log('AndIntLit16 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CCCC, 'I')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpressionLit(Op.AND, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# or-int/lit16 vA, vB, #+CCCC ( 4b, 4b, 16b )
def orintlit16(ins, vmap):
    util.log('OrIntLit16 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CCCC, 'I')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpressionLit(Op.OR, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# xor-int/lit16 vA, vB, #+CCCC ( 4b, 4b, 16b )
def xorintlit16(ins, vmap):
    util.log('XorIntLit16 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CCCC, 'I')
    a, b = get_variables(vmap, ins.A, ins.B)
    exp = BinaryExpressionLit(Op.XOR, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# add-int/lit8 vAA, vBB, #+CC ( 8b, 8b, 8b )
def addintlit8(ins, vmap):
    util.log('AddIntLit8 : %s' % ins.get_output(), 'debug')
    # TODO: generalize this to other operators ?
    if ins.CC < 0:
        literal = -ins.CC
        op = Op.SUB
    else:
        literal = ins.CC
        op = Op.ADD
    cst = Constant(literal, 'I')
    a, b = get_variables(vmap, ins.AA, ins.BB)
    exp = BinaryExpressionLit(op, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# rsub-int/lit8 vAA, vBB, #+CC ( 8b, 8b, 8b )
def rsubintlit8(ins, vmap):
    util.log('RSubIntLit8 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CC, 'I')
    a, b = get_variables(vmap, ins.AA, ins.BB)
    exp = BinaryExpressionLit(Op.SUB, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# mul-int/lit8 vAA, vBB, #+CC ( 8b, 8b, 8b )
def mulintlit8(ins, vmap):
    util.log('MulIntLit8 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CC, 'I')
    a, b = get_variables(vmap, ins.AA, ins.BB)
    exp = BinaryExpressionLit(Op.MUL, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# div-int/lit8 vAA, vBB, #+CC ( 8b, 8b, 8b )
def divintlit8(ins, vmap):
    util.log('DivIntLit8 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CC, 'I')
    a, b = get_variables(vmap, ins.AA, ins.BB)
    exp = BinaryExpressionLit(Op.DIV, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# rem-int/lit8 vAA, vBB, #+CC ( 8b, 8b, 8b )
def remintlit8(ins, vmap):
    util.log('RemIntLit8 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CC, 'I')
    a, b = get_variables(vmap, ins.AA, ins.BB)
    exp = BinaryExpressionLit(Op.MOD, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# and-int/lit8 vAA, vBB, #+CC ( 8b, 8b, 8b )
def andintlit8(ins, vmap):
    util.log('AndIntLit8 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CC, 'I')
    a, b = get_variables(vmap, ins.AA, ins.BB)
    exp = BinaryExpressionLit(Op.ADD, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# or-int/lit8 vAA, vBB, #+CC ( 8b, 8b, 8b )
def orintlit8(ins, vmap):
    util.log('OrIntLit8 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CC, 'I')
    a, b, = get_variables(vmap, ins.AA, ins.BB)
    exp = BinaryExpressionLit(Op.OR, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# xor-int/lit8 vAA, vBB, #+CC ( 8b, 8b, 8b )
def xorintlit8(ins, vmap):
    util.log('XorIntLit8 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CC, 'I')
    a, b, = get_variables(vmap, ins.AA, ins.BB)
    exp = BinaryExpressionLit(Op.XOR, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# shl-int/lit8 vAA, vBB, #+CC ( 8b, 8b, 8b )
def shlintlit8(ins, vmap):
    util.log('ShlIntLit8 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CC, 'I')
    a, b, = get_variables(vmap, ins.AA, ins.BB)
    exp = BinaryExpressionLit(Op.INTSHL, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# shr-int/lit8 vAA, vBB, #+CC ( 8b, 8b, 8b )
def shrintlit8(ins, vmap):
    util.log('ShrIntLit8 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CC, 'I')
    a, b, = get_variables(vmap, ins.AA, ins.BB)
    exp = BinaryExpressionLit(Op.INTSHR, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


# ushr-int/lit8 vAA, vBB, #+CC ( 8b, 8b, 8b )
def ushrintlit8(ins, vmap):
    util.log('UShrIntLit8 : %s' % ins.get_output(), 'debug')
    cst = Constant(ins.CC, 'I')
    a, b, = get_variables(vmap, ins.AA, ins.BB)
    exp = BinaryExpressionLit(Op.INTSHR, b, cst)
    exp.type = 'I'
    return AssignExpression(a, exp)


INSTRUCTION_SET = {
    'nop':                    nop,
    'move':                   move,
    'move/from16':            movefrom16,
    'move/16':                move16,
    'move-wide':              movewide,
    'move-wide/from16':       movewidefrom16,
    'move-wide/16':           movewide16,
    'move-object':            moveobject,
    'move-object/from16':     moveobjectfrom16,
    'move-object/16':         moveobject16,
    'move-result':            moveresult,
    'move-result-wide':       moveresultwide,
    'move-result-object':     moveresultobject,
    'move-exception':         moveexception,
    'return-void':            returnvoid,
    'return':                 return_reg,
    'return-wide':            returnwide,
    'return-object':          returnobject,
    'const/4':                const4,
    'const/16':               const16,
    'const':                  const,
    'const/high16':           consthigh16,
    'const-wide/16':          constwide16,
    'const-wide/32':          constwide32,
    'const-wide':             constwide,
    'const-wide/high16':      constwidehigh16,
    'const-string':           conststring,
    'const-string/jumbo':     conststringjumbo,
    'const-class':            constclass,
    'monitor-enter':          monitorenter,
    'monitor-exit':           monitorexit,
    'check-cast':             checkcast,
    'instance-of':            instanceof,
    'array-length':           arraylength,
    'new-instance':           newinstance,
    'new-array':              newarray,
    'filled-new-array':       fillednewarray,
    'filled-new-array/range': fillednewarrayrange,
    'fill-array-data':        fillarraydata,
    'fill-array-data-payload': fillarraydatapayload,
    'throw':                  throw,
    'goto':                   goto,
    'goto/16':                goto16,
    'goto/32':                goto32,
    'packed-switch':          packedswitch,
    'sparse-switch':          sparseswitch,
    'cmpl-float':             cmplfloat,
    'cmpg-float':             cmpgfloat,
    'cmpl-double':            cmpldouble,
    'cmpg-double':            cmpgdouble,
    'cmp-long':               cmplong,
    'if-eq':                  ifeq,
    'if-ne':                  ifne,
    'if-lt':                  iflt,
    'if-ge':                  ifge,
    'if-gt':                  ifgt,
    'if-le':                  ifle,
    'if-eqz':                 ifeqz,
    'if-nez':                 ifnez,
    'if-ltz':                 ifltz,
    'if-gez':                 ifgez,
    'if-gtz':                 ifgtz,
    'if-lez':                 iflez,
    'aget':                   aget,
    'aget-wide':              agetwide,
    'aget-object':            agetobject,
    'aget-boolean':           agetboolean,
    'aget-byte':              agetbyte,
    'aget-char':              agetchar,
    'aget-short':             agetshort,
    'aput':                   aput,
    'aput-wide':              aputwide,
    'aput-object':            aputobject,
    'aput-boolean':           aputboolean,
    'aput-byte':              aputbyte,
    'aput-char':              aputchar,
    'aput-short':             aputshort,
    'iget':                   iget,
    'iget-wide':              igetwide,
    'iget-object':            igetobject,
    'iget-boolean':           igetboolean,
    'iget-byte':              igetbyte,
    'iget-char':              igetchar,
    'iget-short':             igetshort,
    'iput':                   iput,
    'iput-wide':              iputwide,
    'iput-object':            iputobject,
    'iput-boolean':           iputboolean,
    'iput-byte':              iputbyte,
    'iput-char':              iputchar,
    'iput-short':             iputshort,
    'sget':                   sget,
    'sget-wide':              sgetwide,
    'sget-object':            sgetobject,
    'sget-boolean':           sgetboolean,
    'sget-byte':              sgetbyte,
    'sget-char':              sgetchar,
    'sget-short':             sgetshort,
    'sput':                   sput,
    'sput-wide':              sputwide,
    'sput-object':            sputobject,
    'sput-boolean':           sputboolean,
    'sput-byte':              sputbyte,
    'sput-char':              sputchar,
    'sput-short':             sputshort,
    'invoke-virtual':         invokevirtual,
    'invoke-super':           invokesuper,
    'invoke-direct':          invokedirect,
    'invoke-static':          invokestatic,
    'invoke-interface':       invokeinterface,
    'invoke-virtual/range':   invokevirtualrange,
    'invoke-super/range':     invokesuperrange,
    'invoke-direct/range':    invokedirectrange,
    'invoke-static/range':    invokestaticrange,
    'invoke-interface/range': invokeinterfacerange,
    'neg-int':                negint,
    'not-int':                notint,
    'neg-long':               neglong,
    'not-long':               notlong,
    'neg-float':              negfloat,
    'neg-double':             negdouble,
    'int-to-long':            inttolong,
    'int-to-float':           inttofloat,
    'int-to-double':          inttodouble,
    'long-to-int':            longtoint,
    'long-to-float':          longtofloat,
    'long-to-double':         longtodouble,
    'float-to-int':           floattoint,
    'float-to-long':          floattolong,
    'float-to-double':        floattodouble,
    'double-to-int':          doubletoint,
    'double-to-long':         doubletolong,
    'double-to-float':        doubletofloat,
    'int-to-byte':            inttobyte,
    'int-to-char':            inttochar,
    'int-to-short':           inttoshort,
    'add-int':                addint,
    'sub-int':                subint,
    'mul-int':                mulint,
    'div-int':                divint,
    'rem-int':                remint,
    'and-int':                andint,
    'or-int':                 orint,
    'xor-int':                xorint,
    'shl-int':                shlint,
    'shr-int':                shrint,
    'ushr-int':               ushrint,
    'add-long':               addlong,
    'sub-long':               sublong,
    'mul-long':               mullong,
    'div-long':               divlong,
    'rem-long':               remlong,
    'and-long':               andlong,
    'or-long':                orlong,
    'xor-long':               xorlong,
    'shl-long':               shllong,
    'shr-long':               shrlong,
    'ushr-long':              ushrlong,
    'add-float':              addfloat,
    'sub-float':              subfloat,
    'mul-float':              mulfloat,
    'div-float':              divfloat,
    'rem-float':              remfloat,
    'add-double':             adddouble,
    'sub-double':             subdouble,
    'mul-double':             muldouble,
    'div-double':             divdouble,
    'rem-double':             remdouble,
    'add-int/2addr':          addint2addr,
    'sub-int/2addr':          subint2addr,
    'mul-int/2addr':          mulint2addr,
    'div-int/2addr':          divint2addr,
    'rem-int/2addr':          remint2addr,
    'and-int/2addr':          andint2addr,
    'or-int/2addr':           orint2addr,
    'xor-int/2addr':          xorint2addr,
    'shl-int/2addr':          shlint2addr,
    'shr-int/2addr':          shrint2addr,
    'ushr-int/2addr':         ushrint2addr,
    'add-long/2addr':         addlong2addr,
    'sub-long/2addr':         sublong2addr,
    'mul-long/2addr':         mullong2addr,
    'div-long/2addr':         divlong2addr,
    'rem-long/2addr':         remlong2addr,
    'and-long/2addr':         andlong2addr,
    'or-long/2addr':          orlong2addr,
    'xor-long/2addr':         xorlong2addr,
    'shl-long/2addr':         shllong2addr,
    'shr-long/2addr':         shrlong2addr,
    'ushr-long/2addr':        ushrlong2addr,
    'add-float/2addr':        addfloat2addr,
    'sub-float/2addr':        subfloat2addr,
    'mul-float/2addr':        mulfloat2addr,
    'div-float/2addr':        divfloat2addr,
    'rem-float/2addr':        remfloat2addr,
    'add-double/2addr':       adddouble2addr,
    'sub-double/2addr':       subdouble2addr,
    'mul-double/2addr':       muldouble2addr,
    'div-double/2addr':       divdouble2addr,
    'rem-double/2addr':       remdouble2addr,
    'add-int/lit16':          addintlit16,
    'rsub-int':               rsubint,
    'mul-int/lit16':          mulintlit16,
    'div-int/lit16':          divintlit16,
    'rem-int/lit16':          remintlit16,
    'and-int/lit16':          andintlit16,
    'or-int/lit16':           orintlit16,
    'xor-int/lit16':          xorintlit16,
    'add-int/lit8':           addintlit8,
    'rsub-int/lit8':          rsubintlit8,
    'mul-int/lit8':           mulintlit8,
    'div-int/lit8':           divintlit8,
    'rem-int/lit8':           remintlit8,
    'and-int/lit8':           andintlit8,
    'or-int/lit8':            orintlit8,
    'xor-int/lit8':           xorintlit8,
    'shl-int/lit8':           shlintlit8,
    'shr-int/lit8':           shrintlit8,
    'ushr-int/lit8':          ushrintlit8,
}



#!/usr/bin/env python

# This file is part of Androguard.
#
# Copyright (C) 2012, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

import sys, os

from optparse import OptionParser

from androguard.core import androconf
from androguard.core.bytecodes import apk
from androguard.core.analysis import risk

option_0 = { 'name' : ('-i', '--input'), 'help' : 'file : use these filenames', 'nargs' : 1 }
option_1 = { 'name' : ('-m', '--method'), 'help' : 'perform analysis of each method', 'action' : 'count' }
option_2 = { 'name' : ('-d', '--directory'), 'help' : 'directory : use this directory', 'nargs' : 1 }
option_3 = { 'name' : ('-v', '--version'), 'help' : 'version of the API', 'action' : 'count' }

options = [option_0, option_1, option_2, option_3]

def display_result(res) :
  for i in res :
    print "\t", i
    for j in res[i] :
      print "\t\t", j, res[i][j]

def analyze_app(filename, ri, a) :
    print filename
    display_result( ri.with_apk( a ) )

def analyze_dex(filename, ri, d) :
    print filename
    display_result( ri.with_dex( d ) )

def main(options, arguments) :
    ri = risk.RiskIndicator()
    ri.add_risk_analysis( risk.RedFlags() )
    ri.add_risk_analysis( risk.FuzzyRisk() )

    if options.input != None :
        ret_type = androconf.is_android( options.input ) 
        if ret_type == "APK" :
            a = apk.APK( options.input )
            analyze_app( options.input, ri, a )
        elif ret_type == "DEX" :
            analyze_dex( options.input, ri, open(options.input, "r").read() )


    elif options.directory != None :
        for root, dirs, files in os.walk( options.directory, followlinks=True ) :
            if files != [] :
                for f in files :
                    real_filename = root
                    if real_filename[-1] != "/" :
                        real_filename += "/"
                    real_filename += f

                    ret_type = androconf.is_android( real_filename )
                    if ret_type == "APK"  :
                        try :
                            a = apk.APK( real_filename )
                            analyze_app( real_filename, ri, a )
                        except Exception, e :
                            print e

                    elif ret_type == "DEX" :
                        analyze_dex( real_filename, ri, open(real_filename, "r").read() )

    elif options.version != None :
        print "Androrisk version %s" % androconf.ANDROGUARD_VERSION

if __name__ == "__main__" :
    parser = OptionParser()
    for option in options :
        param = option['name']
        del option['name']
        parser.add_option(*param, **option)

    options, arguments = parser.parse_args()
    sys.argv[:] = arguments
    main(options, arguments)



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name = 'vandroguard',
    version = '1.5.1',
    packages = find_packages(),
    scripts = ['androaxml.py', 'androcsign.py', 'androdiff.py', 'androgexf.py',
               'androlyze.py', 'andromercury.py', 'androrisk.py', 'androsign.py',
               'androsim.py', 'androxgmml.py', 'apkviewer.py',
               'androdd.py', 'androapkinfo.py',
               ],
    install_requires=['distribute'],
)



#!/usr/bin/env python

# This file is part of Androguard.
#
# Copyright (C) 2012, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

import sys, re, os

from optparse import OptionParser

from androguard.core.bytecodes import apk

sys.path.append("./elsim/")
from elsim.elsign import dalvik_elsign

sys.path.append("./mercury/client")
from merc.lib.common import Session

option_0 = { 'name' : ('-l', '--list'), 'help' : 'list all packages', 'nargs' : 1 }
option_1 = { 'name' : ('-i', '--input'), 'help' : 'get specific packages (a filter)', 'nargs' : 1 }
option_2 = { 'name' : ('-r', '--remotehost'), 'help' : 'specify ip of emulator/device', 'nargs' : 1 }
option_3 = { 'name' : ('-p', '--port'), 'help' : 'specify the port', 'nargs' : 1 }
option_4 = { 'name' : ('-o', '--output'), 'help' : 'output directory to write packages', 'nargs' : 1 }
option_5 = { 'name' : ('-b', '--database'), 'help' : 'database : use this database', 'nargs' : 1 }
option_6 = { 'name' : ('-c', '--config'), 'help' : 'use this configuration', 'nargs' : 1 }
option_7 = { 'name' : ('-v', '--verbose'), 'help' : 'display debug information', 'action' : 'count' }

options = [option_0, option_1, option_2, option_3, option_4, option_5, option_6, option_7]

def display(ret, debug) :
    print "---->", ret[0],

def main(options, arguments) :
    sessionip = "127.0.0.1"
    sessionport = 31415
    
    if options.remotehost :
        sessionip = options.remotehost

    if options.port :
        sessionport = int(options.port)

    newsession = Session(sessionip, sessionport, "bind")

    # Check if connection can be established
    if newsession.executeCommand("core", "ping", None).data == "pong":
      
        if options.list :
            request = {'filter': options.list, 'permissions': None }
            apks_info = newsession.executeCommand("packages", "info", {}).getPaddedErrorOrData()
            print apks_info

        elif options.input and options.output :
            s = None
            if options.database != None or options.config != None :
                s = dalvik_elsign.MSignature( options.database, options.config, options.verbose != None, ps = dalvik_elsign.PublicSignature)
            
            request = {'filter': options.input, 'permissions': None }
            apks_info = newsession.executeCommand("packages", "info", request).getPaddedErrorOrData()
            print apks_info

            for i in apks_info.split("\n") :
                if re.match("APK path:", i) != None :
                    name_app = i.split(":")[1][1:]
                    print name_app,
                    response = newsession.downloadFile(name_app, options.output)
                    print response.data, response.error,
                    
                    if s != None :
                        a = apk.APK( options.output + "/" + os.path.basename(name_app) )
                        if a.is_valid_APK() :
                            display( s.check_apk( a ), options.verbose )
                    print
    else:
        print "\n**Network Error** Could not connect to " + sessionip + ":" + str(sessionport) + "\n"

if __name__ == "__main__" :
    parser = OptionParser()
    for option in options :
        param = option['name']
        del option['name']
        parser.add_option(*param, **option)

    options, arguments = parser.parse_args()
    sys.argv[:] = arguments
    main(options, arguments)



# This file is part of Androguard.
#
# Copyright (C) 2012, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

from elfesteem import *
from miasm.tools.pe_helper import *
from miasm.core import asmbloc
from miasm.arch import arm_arch
from miasm.core import bin_stream


from androguard.core import bytecode
from androguard.core.androconf import CONF, debug

def disasm_at_addr(in_str, ad_to_dis, symbol_pool) :
    kargs = {}
    all_bloc = asmbloc.dis_bloc_all(arm_arch.arm_mn, in_str, ad_to_dis, set(),
                                        symbol_pool=symbol_pool,
                                        dontdis_retcall = False,
                                        follow_call = False,
                                        **kargs)
    for i in all_bloc :
        bytecode._PrintDefault("%s\n" % i.label)
        for j in i.lines :
            bytecode._PrintDefault("\t %s\n" % j)
        bytecode._PrintDefault("\n")

class Function :
    def __init__(self, cm, name, info) :
        self.cm = cm
        self.name = name
        self.info = info

    def show(self) :
        bytecode._PrintSubBanner("Function")
        bytecode._PrintDefault("name=%s addr=0x%x\n" % (self.name, self.info.value))
        
        self.cm.disasm_at_addr( self.info.value )

class ClassManager :
    def __init__(self, in_str, symbol_pool) :
        self.in_str = in_str
        self.symbol_pool = symbol_pool

    def disasm_at_addr(self, ad_to_dis) :
        disasm_at_addr( self.in_str, ad_to_dis, self.symbol_pool )

class ELF :
    def __init__(self, buff) :
        self.E = elf_init.ELF( buff )

        self.in_str = bin_stream.bin_stream(self.E.virt)
        self.symbol_pool = None
        self.functions = []

        self.create_symbol_pool()

        self.CM = ClassManager( self.in_str, self.symbol_pool )
        
        self.create_functions()

    def create_symbol_pool(self) :
        dll_dyn_funcs = get_import_address_elf(self.E)
        self.symbol_pool = asmbloc.asm_symbol_pool()
        for (n,f), ads in dll_dyn_funcs.items() :
            for ad in ads :
                l  = self.symbol_pool.getby_name_create("%s_%s"%(n, f))
                l.offset = ad
                self.symbol_pool.s_offset[l.offset] = l

    def show(self) :
        for i in self.get_functions():
            i.show()

    def get_functions(self) :
        return self.functions

    def create_functions(self) :
        try :
            for k, v in self.E.sh.symtab.symbols.items():
                if v.size != 0 :
                    self.functions.append( Function(self.CM, k, v) )
        except AttributeError :
            pass
        
        for k, v in self.E.sh.dynsym.symbols.items() :
            if v.size != 0 :
                self.functions.append( Function(self.CM, k, v) )



#!/usr/bin/env python

# This file is part of Androguard.
#
# Copyright (C) 2012, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

import sys
from optparse import OptionParser
from xml.dom import minidom
import codecs

from androguard.core import androconf
from androguard.core.bytecodes import apk


option_0 = { 'name' : ('-i', '--input'), 'help' : 'filename input (APK or android\'s binary xml)', 'nargs' : 1 }
option_1 = { 'name' : ('-o', '--output'), 'help' : 'filename output of the xml', 'nargs' : 1 }
option_2 = { 'name' : ('-v', '--version'), 'help' : 'version of the API', 'action' : 'count' }
options = [option_0, option_1, option_2]

def main(options, arguments) :
    if options.input != None :
        buff = ""

        ret_type = androconf.is_android( options.input )
        if ret_type == "APK" :
            a = apk.APK( options.input )
            buff = a.xml[ "AndroidManifest.xml" ].toprettyxml()
            a.get_activities()
        elif ".xml" in options.input :
            ap = apk.AXMLPrinter( open(options.input, "rb").read() )
            buff = minidom.parseString( ap.getBuff() ).toprettyxml()
        else :
            print "Unknown file type"
            return

        if options.output != None :
            fd = codecs.open(options.output, "w", "utf-8")
            fd.write( buff )
            fd.close()
        else :
            print buff

    elif options.version != None :
        print "Androaxml version %s" % androconf.ANDROGUARD_VERSION

if __name__ == "__main__" :
    parser = OptionParser()
    for option in options :
        param = option['name']
        del option['name']
        parser.add_option(*param, **option)

    options, arguments = parser.parse_args()
    sys.argv[:] = arguments
    main(options, arguments)



#!/usr/bin/env python

# This file is part of Androguard.
#
# Copyright (C) 2012, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, cmd, threading, code, re


from optparse import OptionParser

from androguard.core import *
from androguard.core.androgen import *
from androguard.core.androconf import *
from androguard.core.bytecode import *
from androguard.core.bytecodes.jvm import *
from androguard.core.bytecodes.dvm import *
from androguard.core.bytecodes.apk import *

from androguard.core.analysis.analysis import *
from androguard.core.analysis.ganalysis import *
from androguard.core.analysis.risk import *

from androguard.decompiler.decompiler import *

from androguard.core import androconf

from IPython.frontend.terminal.embed import InteractiveShellEmbed
from IPython.config.loader import Config

from cPickle import dumps, loads

option_0 = { 'name' : ('-i', '--input'), 'help' : 'file : use this filename', 'nargs' : 1 }
option_1 = { 'name' : ('-d', '--display'), 'help' : 'display the file in human readable format', 'action' : 'count' }
option_2 = { 'name' : ('-m', '--method'), 'help' : 'display method(s) respect with a regexp', 'nargs' : 1 }
option_3 = { 'name' : ('-f', '--field'), 'help' : 'display field(s) respect with a regexp', 'nargs' : 1 }
option_4 = { 'name' : ('-s', '--shell'), 'help' : 'open an interactive shell to play more easily with objects', 'action' : 'count' }
option_5 = { 'name' : ('-v', '--version'), 'help' : 'version of the API', 'action' : 'count' }
option_6 = { 'name' : ('-p', '--pretty'), 'help' : 'pretty print !', 'action' : 'count' }
option_8 = { 'name' : ('-x', '--xpermissions'), 'help' : 'show paths of permissions', 'action' : 'count' }

options = [option_0, option_1, option_2, option_3, option_4, option_5, option_6, option_8]

def init_print_colors() :
    from IPython.utils import coloransi, io
    default_colors(coloransi.TermColors)
    CONF["PRINT_FCT"] = io.stdout.write

def interact() :
    cfg = Config()
    ipshell = InteractiveShellEmbed(config=cfg, banner1="Androlyze version %s" % androconf.ANDROGUARD_VERSION)
    init_print_colors()
    ipshell()

def save_session(l, filename) :
    """
        save your session !

        :param l: a list of objects
        :type: a list of object
        :param filename: output filename to save the session
        :type filename: string

        :Example:
            save_session([a, vm, vmx], "msession.json")
    """
    fd = open(filename, "w")
    fd.write( dumps(l, -1) )
    fd.close()

def load_session(filename) :
    """
        load your session !

        :param filename: the filename where the session has been saved
        :type filename: string
        
        :rtype: the elements of your session :)

        :Example: 
            a, vm, vmx = load_session("mysession.json")
    """
    return loads( open(filename, "r").read() )

def AnalyzeAPK(filename, raw=False, decompiler=None) :
    """
        Analyze an android application and setup all stuff for a more quickly analysis !

        :param filename: the filename of the android application or a buffer which represents the application
        :type filename: string
        :param raw: True is you would like to use a buffer (optional)
        :type raw: boolean
        :param decompiler: ded, dex2jad, dad (optional)
        :type decompiler: string
        
        :rtype: return the :class:`APK`, :class:`DalvikVMFormat`, and :class:`VMAnalysis` objects
    """
    androconf.debug("APK ...")
    a = APK(filename, raw)

    d, dx = AnalyzeDex( a.get_dex(), raw=True, decompiler=decompiler )

    return a, d, dx


def AnalyzeDex(filename, raw=False, decompiler=None) :
    """
        Analyze an android dex file and setup all stuff for a more quickly analysis !

        :param filename: the filename of the android dex file or a buffer which represents the dex file
        :type filename: string
        :param raw: True is you would like to use a buffer (optional)
        :type raw: boolean

        :rtype: return the :class:`DalvikVMFormat`, and :class:`VMAnalysis` objects
    """
    androconf.debug("DalvikVMFormat ...")
    d = None
    if raw == False :
        d = DalvikVMFormat( open(filename, "rb").read() )
    else :
        d = DalvikVMFormat( filename )

    androconf.debug("Export VM to python namespace")
    d.create_python_export()

    androconf.debug("VMAnalysis ...")
    dx = uVMAnalysis( d )

    androconf.debug("GVMAnais ...")
    gx = GVMAnalysis( dx, None )

    d.set_vmanalysis( dx )
    d.set_gvmanalysis( gx )

    RunDecompiler( d, dx, decompiler )

    androconf.debug("XREF ...")
    d.create_xref()
    androconf.debug("DREF ...")
    d.create_dref()

    return d, dx

def RunDecompiler(d, dx, decompiler) :
    """
        Run the decompiler on a specific analysis

        :param d: the DalvikVMFormat object
        :type d: :class:`DalvikVMFormat` object
        :param dx: the analysis of the format
        :type dx: :class:`VMAnalysis` object 
        :param decompiler: the type of decompiler to use ("dad", "dex2jad", "ded")
        :type decompiler: string
    """
    if decompiler != None :
      androconf.debug("Decompiler ...")
      decompiler = decompiler.lower()
      if decompiler == "dex2jad" :
        d.set_decompiler( DecompilerDex2Jad( d, androconf.CONF["PATH_DEX2JAR"], androconf.CONF["BIN_DEX2JAR"], androconf.CONF["PATH_JAD"], androconf.CONF["BIN_JAD"] ) )
      elif decompiler == "ded" :
        d.set_decompiler( DecompilerDed( d, androconf.CONF["PATH_DED"], androconf.CONF["BIN_DED"] ) )
      elif decompiler == "dad" :
        d.set_decompiler( DecompilerDAD( d, dx ) )
      else :
        print "Unknown decompiler, use DAD decompiler by default"
        d.set_decompiler( DecompilerDAD( d, dx ) )

def AnalyzeElf(filename, raw=False) :
    # avoid to install smiasm for everybody
    from androguard.core.binaries.elf import ELF 

    e = None
    if raw == False:
        e = ELF( open(filename, "rb").read() )
    else:
        e = ELF( filename )

    ExportElfToPython( e )

    return e

def ExportElfToPython(e) :
    for function in e.get_functions():
        name = "FUNCTION_" + function.name
        setattr( e, name, function )
        
def AnalyzeJAR(filename, raw=False) :
    androconf.debug("JAR ...")
    a = JAR(filename, raw)

    d = AnalyzeClasses( a.get_classes() )

    return a, d

def AnalyzeClasses( classes ) :
  d = {}
  for i in classes :
    d[i[0]] = JVMFormat( i[1] )

  return d

def main(options, arguments) :
    if options.shell != None :
        interact()

    elif options.input != None :
        _a = AndroguardS( options.input )

        if options.pretty != None :
          init_print_colors()

        if options.display != None :
            if options.pretty != None :
                _a.ianalyze()
                _a.pretty_show()
            else :
                _a.show()

        elif options.method != None :
            for method in _a.get("method", options.method) :
                if options.pretty != None :
                    _a.ianalyze()
                    method.pretty_show() 
                else :
                    method.show()

        elif options.field != None :
            for field in _a.get("field", options.field) :
                field.show()

        elif options.xpermissions != None :
            _a.ianalyze()
            perms_access = _a.get_analysis().get_permissions( [] )
            for perm in perms_access :
                print "PERM : ", perm
                for path in perms_access[ perm ] :
                    show_Path( _a.get_vm(), path )

    elif options.version != None :
        print "Androlyze version %s" % androconf.ANDROGUARD_VERSION

if __name__ == "__main__" :
    parser = OptionParser()
    for option in options :
        param = option['name']
        del option['name']
        parser.add_option(*param, **option)

    options, arguments = parser.parse_args()
    sys.argv[:] = arguments
    main(options, arguments)



# This file is part of Androguard.
#
# Copyright (C) 2011, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

from subprocess import Popen, PIPE, STDOUT

import tempfile, os

from androguard.core.androconf import rrmdir

from pygments.filter import Filter
from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import HtmlFormatter, TerminalFormatter
from pygments.token import Token, Text, STANDARD_TYPES

class DecompilerDex2Jad :
    def __init__(self, vm, path_dex2jar = "./decompiler/dex2jar/", bin_dex2jar = "dex2jar.sh", path_jad="./decompiler/jad/", bin_jad="jad") :
        self.classes = {}
        self.classes_failed = []
        
        pathtmp = os.getcwd() + "/tmp/"
        if not os.path.exists(pathtmp) :
            os.makedirs( pathtmp )

        fd, fdname = tempfile.mkstemp( dir=pathtmp )
        fd = os.fdopen(fd, "w+b")
        fd.write( vm.get_buff() )
        fd.flush()
        fd.close()
       
        compile = Popen([ path_dex2jar + bin_dex2jar, fdname ], stdout=PIPE, stderr=STDOUT)        
        stdout, stderr = compile.communicate()
        os.unlink( fdname )

        pathclasses = fdname + "dex2jar/"
        compile = Popen([ "unzip", fdname + "_dex2jar.jar", "-d", pathclasses ], stdout=PIPE, stderr=STDOUT)        
        stdout, stderr = compile.communicate()
        os.unlink( fdname + "_dex2jar.jar" )

        for root, dirs, files in os.walk( pathclasses, followlinks=True ) :
            if files != [] :
                for f in files :
                    real_filename = root
                    if real_filename[-1] != "/" :
                        real_filename += "/"
                    real_filename += f
                    
                    compile = Popen([ path_jad + bin_jad, "-o", "-d", root, real_filename ], stdout=PIPE, stderr=STDOUT)
                    stdout, stderr = compile.communicate()

        for i in vm.get_classes() :
            fname = pathclasses + "/" + i.get_name()[1:-1] + ".jad"
            if os.path.isfile(fname) == True :
                fd = open(fname, "r")
                self.classes[ i.get_name() ] = fd.read() 
                fd.close()
            else :
                self.classes_failed.append( i.get_name() )
    
        rrmdir( pathclasses )

    def get_source(self, class_name, method_name) :
        if class_name not in self.classes :
            return ""

        lexer = get_lexer_by_name("java", stripall=True)
        lexer.add_filter(MethodFilter(method_name=method_name))
        formatter = TerminalFormatter()
        result = highlight(self.classes[class_name], lexer, formatter)
        return result

    def display_source(self, class_name, method_name, method_descriptor) :
        print self.get_source( class_name, method_name )

    def get_all(self, class_name) :
        if class_name not in self.classes :
            return ""

        lexer = get_lexer_by_name("java", stripall=True)
        formatter = TerminalFormatter()
        result = highlight(self.classes[class_name], lexer, formatter)
        
        return result

    def display_all(self, class_name) :
        print self.get_all( class_name )

class DecompilerDed :
    def __init__(self, vm, path="./decompiler/ded/", bin_ded = "ded.sh") :
        self.classes = {}
        self.classes_failed = []

        pathtmp = os.getcwd() + "/tmp/"
        if not os.path.exists(pathtmp) :
            os.makedirs( pathtmp )

        fd, fdname = tempfile.mkstemp( dir=pathtmp )
        fd = os.fdopen(fd, "w+b")
        fd.write( vm.get_buff() )
        fd.flush()
        fd.close()
       
        dirname = tempfile.mkdtemp(prefix=fdname + "-src")
        compile = Popen([ path + bin_ded, "-c", "-o", "-d", dirname, fdname ], stdout=PIPE, stderr=STDOUT)        
        stdout, stderr = compile.communicate()
        os.unlink( fdname )

        findsrc = None 
        for root, dirs, files in os.walk( dirname + "/optimized-decompiled/" ) :
            if dirs != [] :
                for f in dirs :
                    if f == "src" :
                        findsrc = root
                        if findsrc[-1] != "/" :
                            findsrc += "/"
                        findsrc += f
                        break
            if findsrc != None :
                break
        
        for i in vm.get_classes() :
            fname = findsrc + "/" + i.get_name()[1:-1] + ".java"
            #print fname
            if os.path.isfile(fname) == True :
                fd = open(fname, "r")
                self.classes[ i.get_name() ] = fd.read() 
                fd.close()
            else :
                self.classes_failed.append( i.get_name() )
      
        rrmdir( dirname )

    def get_source(self, class_name, method_name) :
        if class_name not in self.classes :
            return ""

        lexer = get_lexer_by_name("java", stripall=True)
        lexer.add_filter(MethodFilter(method_name=method_name))
        formatter = TerminalFormatter()
        result = highlight(self.classes[class_name], lexer, formatter)
        return result

    def display_source(self, class_name, method_name, method_descriptor) :
        print self.get_source( class_name, method_name )

    def get_all(self, class_name) :
        if class_name not in self.classes :
            return ""

        lexer = get_lexer_by_name("java", stripall=True)
        formatter = TerminalFormatter()
        result = highlight(self.classes[class_name], lexer, formatter)
        return result
    
    def display_all(self, class_name) :
        print self.get_all( class_name )

class MethodFilter(Filter):
    def __init__(self, **options):
        Filter.__init__(self, **options)
        
        self.method_name = options["method_name"]
        #self.descriptor = options["descriptor"]

        self.present = False
        self.get_desc = True #False

    def filter(self, lexer, stream) :
        a = []
        l = []
        rep = []

        for ttype, value in stream:
            if self.method_name == value and (ttype is Token.Name.Function or ttype is Token.Name) :
                #print ttype, value

                item_decl = -1
                for i in range(len(a)-1, 0, -1) :
                    if a[i][0] is Token.Keyword.Declaration :
                        if a[i][1] != "class" :
                            item_decl = i
                        break
               
                if item_decl != -1 :
                    self.present = True
                    l.extend( a[item_decl:] )
        

            if self.present and ttype is Token.Keyword.Declaration :
                item_end = -1
                for i in range(len(l)-1, 0, -1) :
                    if l[i][0] is Token.Operator and l[i][1] == "}" :
                        item_end = i
                        break
               
                if item_end != -1 :
                    rep.extend( l[:item_end+1] )
                    l = []
                    self.present = False
                
            if self.present :
               # if self.get_desc == False :
               #     if ttype is Token.Operator and value == ")" :
               #         self.get_desc = True
               #         
               #         item_desc = -1
               #         for i in range(len(l)-1, 0, -1) :
               #             if l[i][1] == "(" :
               #                 item_desc = i
               #                 break
               #         desc = ''.join(i[1] for i in l[item_desc+1:-1] )
               #         desc = desc.split()
               #         print "DESC", desc, 
                        
               #         equivalent = True

               #         if len(self.descriptor) != len(desc) :
               #             equivalent = False
               #         else :
               #             for i in range(0, len(self.descriptor)) :
               #                 if self.descriptor[i] != desc[i] :
               #                     equivalent = False
               #                     break

               #         print equivalent
               #         if equivalent == False :
               #             self.get_desc = False
               #             self.present = False
               #             l = []

                #print "ADD", (ttype, value) 
                l.append( (ttype, value) )

            a.append( (ttype, value) )

        
        if self.present :
            nb = 0
            item_end = -1
            for i in range(len(l)-1, 0, -1) :
                if l[i][0] is Token.Operator and l[i][1] == "}" :
                    nb += 1
                    if nb == 2 :
                        item_end = i
                        break
            
            rep.extend( l[:item_end+1] )
            #return l[:item_end+1]
            
        return rep

class DecompilerDAD :
    def __init__(self, vm, vmx) :
        self.vm = vm
        self.vmx = vmx

    def display_source(self, m) :
        mx = self.vmx.get_method( m )
        
        from dad import decompile 

        z = decompile.DvMethod( mx )
        z.process()

        lexer = get_lexer_by_name("java", stripall=True)
        formatter = TerminalFormatter()
        result = highlight(z.get_source(), lexer, formatter)
        print result
    
    def display_all(self, _class) :
        from dad import decompile

        c = decompile.DvClass( _class, self.vmx )
        c.process()

        c.show_source()

        #lexer = get_lexer_by_name("java", stripall=True)
        #formatter = TerminalFormatter()
        #result = highlight(c.show_source(), lexer, formatter)
        #print result

    def get_all(self, class_name) :
        pass
